@@ -1,10 +1,9 @@
-# Copyright (c) 2009-2010 Denis Bilenko. See LICENSE for details.
+# Copyright (c) 2009-2012 Denis Bilenko. See LICENSE for details.
 
 import sys
-import traceback
-from gevent import core
-from gevent.hub import greenlet, getcurrent, get_hub, GreenletExit, Waiter, kill
+from gevent.hub import greenlet, getcurrent, get_hub, GreenletExit, Waiter
 from gevent.timeout import Timeout
+from gevent.six import callable, _meth_self, moves, PY3
 
 
 __all__ = ['Greenlet',
@@ -20,6 +19,8 @@ class SpawnedLink(object):
     __slots__ = ['callback']
 
     def __init__(self, callback):
+        if not callable(callback):
+            raise TypeError("Expected callable: %r" % (callback, ))
         self.callback = callback
 
     def __call__(self, source):
@@ -67,98 +68,39 @@ class FailureSpawnedLink(SpawnedLink):
             return SpawnedLink.__call__(self, source)
 
 
-class GreenletLink(object):
-    """A wrapper around greenlet that raises a LinkedExited exception when called.
-
-    Can be called only from main loop.
-    """
-    __slots__ = ['greenlet']
-
-    def __init__(self, greenlet):
-        self.greenlet = greenlet
-
-    def __call__(self, source):
-        if source.successful():
-            if isinstance(source.value, GreenletExit):
-                error = LinkedKilled(source)
-            else:
-                error = LinkedCompleted(source)
-        else:
-            error = LinkedFailed(source)
-        current = getcurrent()
-        greenlet = self.greenlet
-        if current is greenlet:
-            greenlet.throw(error)
-        elif current is get_hub():
-            try:
-                greenlet.throw(error)
-            except:
-                traceback.print_exc()
-        else:
-            kill(self.greenlet, error)
-
-    def __hash__(self):
-        return hash(self.greenlet)
-
-    def __eq__(self, other):
-        return self.greenlet == getattr(other, 'greenlet', other)
-
-    def __str__(self):
-        return str(self.greenlet)
-
-    def __repr__(self):
-        return repr(self.greenlet)
-
-
-class SuccessGreenletLink(GreenletLink):
-    """A wrapper around greenlet that raises a LinkedExited exception when called
-    if source has succeed.
-
-    Can be called only from main loop.
-    """
-    __slots__ = []
-
-    def __call__(self, source):
-        if source.successful():
-            return GreenletLink.__call__(self, source)
-
-
-class FailureGreenletLink(GreenletLink):
-    """A wrapper around greenlet that raises a LinkedExited exception when called
-    if source has failed.
-
-    Can be called only from main loop.
-    """
-    __slots__ = []
-
-    def __call__(self, source):
-        if not source.successful():
-            return GreenletLink.__call__(self, source)
-
-
 class Greenlet(greenlet):
     """A light-weight cooperatively-scheduled execution unit."""
 
-    args = ()
-    kwargs = {}
-
     def __init__(self, run=None, *args, **kwargs):
-        greenlet.__init__(self, parent=get_hub())
+        hub = get_hub()
+        greenlet.__init__(self, parent=hub)
         if run is not None:
             self._run = run
-        if args:
-            self.args = args
-        if kwargs:
-            self.kwargs = kwargs
-        self._links = set()
+        self.args = args
+        self.kwargs = kwargs
+        self._links = []
         self.value = None
         self._exception = _NONE
-        self._notifier = None
+        loop = hub.loop
+        self._notifier = loop.callback()
         self._start_event = None
 
+    @property
+    def loop(self):
+        # needed by killall
+        return self.parent.loop
+
+    if PY3:
+        def __bool__(self):
+            return self._start_event is not None and self._exception is _NONE
+    else:
+        def __nonzero__(self):
+            return self._start_event is not None and self._exception is _NONE
+
     @property
     def started(self):
-        return self._start_event is not None or bool(self)
+        # DEPRECATED
+        return bool(self)
 
     def ready(self):
         """Return true if and only if the greenlet has finished execution."""
@@ -217,17 +159,18 @@ class Greenlet(greenlet):
         a) cancel the event that will start it
         b) fire the notifications as if an exception was raised in a greenlet
         """
-        if self._start_event is not None:
-            self._start_event.cancel()
-            self._start_event = None
+        if self._start_event is None:
+            self._start_event = _dummy_event
+        else:
+            self._start_event.stop()
         try:
             greenlet.throw(self, *args)
         finally:
             if self._exception is _NONE and self.dead:
-                # the greenlet was not started yet, so _report_error was not called, so
+                # the greenlet was never switched to before and it will never be, _report_error was not called
                 # the result was not set and the links weren't notified. let's do it here.
-                # checking that self.dead is true is essential, because the exception raised by
-                # throw() could have been cancelled by the greenlet's function.
+                # checking that self.dead is true is essential, because throw() does not necessarily kill the greenlet
+                # (if the exception raised by throw() is caught somewhere inside the greenlet).
                 if len(args) == 1:
                     arg = args[0]
                     #if isinstance(arg, type):
@@ -241,13 +184,15 @@ class Greenlet(greenlet):
 
     def start(self):
         """Schedule the greenlet to run in this loop iteration"""
-        assert not self.started, 'Greenlet already started'
-        self._start_event = core.active_event(self.switch)
+        if self._start_event is None:
+            self._start_event = self.parent.loop.callback()
+            self._start_event.start(self.switch)
 
     def start_later(self, seconds):
         """Schedule the greenlet to run in the future loop iteration *seconds* later"""
-        assert not self.started, 'Greenlet already started'
-        self._start_event = core.timer(seconds, self.switch)
+        if self._start_event is None:
+            self._start_event = self.parent.loop.timer(seconds)
+            self._start_event.start(self.switch)
 
     @classmethod
     def spawn(cls, *args, **kwargs):
@@ -269,24 +214,6 @@ class Greenlet(greenlet):
         g.start_later(seconds)
         return g
 
-    @classmethod
-    def spawn_link(cls, *args, **kwargs):
-        g = cls.spawn(*args, **kwargs)
-        g.link()
-        return g
-
-    @classmethod
-    def spawn_link_value(cls, *args, **kwargs):
-        g = cls.spawn(*args, **kwargs)
-        g.link_value()
-        return g
-
-    @classmethod
-    def spawn_link_exception(cls, *args, **kwargs):
-        g = cls.spawn(*args, **kwargs)
-        g.link_exception()
-        return g
-
     def kill(self, exception=GreenletExit, block=True, timeout=None):
         """Raise the exception in the greenlet.
 
@@ -297,12 +224,13 @@ class Greenlet(greenlet):
 
         `Changed in version 0.13.0:` *block* is now ``True`` by default.
         """
-        if self._start_event is not None:
-            self._start_event.cancel()
-            self._start_event = None
+        if self._start_event is None:
+            self._start_event = _dummy_event
+        else:
+            self._start_event.stop()
         if not self.dead:
             waiter = Waiter()
-            core.active_event(_kill, self, exception, waiter)
+            self.parent.loop.run_callback(_kill, self, exception, waiter)
             if block:
                 waiter.get()
                 self.join(timeout)
@@ -327,7 +255,7 @@ class Greenlet(greenlet):
             try:
                 t = Timeout.start_new(timeout)
                 try:
-                    result = get_hub().switch()
+                    result = self.parent.switch()
                     assert result is self, 'Invalid switch into Greenlet.get(): %r' % (result, )
                 finally:
                     t.cancel()
@@ -359,13 +287,13 @@ class Greenlet(greenlet):
             try:
                 t = Timeout.start_new(timeout)
                 try:
-                    result = get_hub().switch()
+                    result = self.parent.switch()
                     assert result is self, 'Invalid switch into Greenlet.join(): %r' % (result, )
                 finally:
                     t.cancel()
-            except Timeout, ex:
+            except Timeout:
                 self.unlink(switch)
-                if ex is not t:
+                if sys.exc_info()[1] is not t:
                     raise
             except:
                 self.unlink(switch)
@@ -374,33 +302,27 @@ class Greenlet(greenlet):
     def _report_result(self, result):
         self._exception = None
         self.value = result
-        if self._links and self._notifier is None:
-            self._notifier = core.active_event(self._notify_links)
+        if self._links and not self._notifier.active:
+            self._notifier.start(self._notify_links)
 
     def _report_error(self, exc_info):
         exception = exc_info[1]
         if isinstance(exception, GreenletExit):
             self._report_result(exception)
             return
-        try:
-            traceback.print_exception(*exc_info)
-        except:
-            pass
         self._exception = exception
 
-        if self._links and self._notifier is None:
-            self._notifier = core.active_event(self._notify_links)
+        if self._links and not self._notifier.active:
+            self._notifier.start(self._notify_links)
 
-        info = str(self) + ' failed with '
-        try:
-            info += self._exception.__class__.__name__
-        except Exception:
-            info += str(self._exception) or repr(self._exception)
-        sys.stderr.write(info + '\n\n')
+        self.parent.handle_error(self, *exc_info)
 
     def run(self):
         try:
-            self._start_event = None
+            if self._start_event is None:
+                self._start_event = _dummy_event
+            else:
+                self._start_event.stop()
             try:
                 result = self._run(*self.args, **self.kwargs)
             except:
@@ -419,104 +341,85 @@ class Greenlet(greenlet):
         """
         if not callable(callback):
             raise TypeError('Expected callable: %r' % (callback, ))
-        self._links.add(callback)
-        if self.ready() and self._notifier is None:
-            self._notifier = core.active_event(self._notify_links)
+        self._links.append(callback)
+        if self.ready() and not self._notifier.active:
+            self._notifier.start(self._notify_links)
 
-    def link(self, receiver=None, GreenletLink=GreenletLink, SpawnedLink=SpawnedLink):
-        """Link greenlet's completion to callable or another greenlet.
+    def link(self, receiver, SpawnedLink=SpawnedLink):
+        """Link greenlet's completion to a callable.
 
-        If *receiver* is a callable then it will be called with this instance as an argument
+        The *receiver* will be called with this instance as an argument
         once this greenlet's dead. A callable is called in its own greenlet.
-
-        If *receiver* is a greenlet then an :class:`LinkedExited` exception will be
-        raised in it once this greenlet's dead.
-
-        If *receiver* is ``None``, link to the current greenlet.
-
-        Always asynchronous, unless receiver is a current greenlet and the result is ready.
-        If this greenlet is already dead, then notification will performed in this loop
-        iteration as soon as this greenlet switches to the hub.
         """
-        current = getcurrent()
-        if receiver is None or receiver is current:
-            receiver = GreenletLink(current)
-            if self.ready():
-                # special case : linking to current greenlet when the result is ready
-                # raise LinkedExited immediatelly
-                receiver(self)
-                return
-        elif not callable(receiver):
-            if isinstance(receiver, greenlet):
-                receiver = GreenletLink(receiver)
-            else:
-                raise TypeError('Expected callable or greenlet: %r' % (receiver, ))
-        else:
-            receiver = SpawnedLink(receiver)
-        self.rawlink(receiver)
+        self.rawlink(SpawnedLink(receiver))
 
-    def unlink(self, receiver=None):
+    def unlink(self, receiver):
         """Remove the receiver set by :meth:`link` or :meth:`rawlink`"""
-        if receiver is None:
-            receiver = getcurrent()
-        # discarding greenlets when we have GreenletLink instances in _links works, because
-        # a GreenletLink instance pretends to be a greenlet, hash-wise and eq-wise
-        self._links.discard(receiver)
+        try:
+            self._links.remove(receiver)
+        except ValueError:
+            pass
 
-    def link_value(self, receiver=None, GreenletLink=SuccessGreenletLink, SpawnedLink=SuccessSpawnedLink):
+    def link_value(self, receiver, SpawnedLink=SuccessSpawnedLink):
         """Like :meth:`link` but *receiver* is only notified when the greenlet has completed successfully"""
-        self.link(receiver=receiver, GreenletLink=GreenletLink, SpawnedLink=SpawnedLink)
+        self.link(receiver=receiver, SpawnedLink=SpawnedLink)
 
-    def link_exception(self, receiver=None, GreenletLink=FailureGreenletLink, SpawnedLink=FailureSpawnedLink):
+    def link_exception(self, receiver=None, SpawnedLink=FailureSpawnedLink):
         """Like :meth:`link` but *receiver* is only notified when the greenlet dies because of unhandled exception"""
-        self.link(receiver=receiver, GreenletLink=GreenletLink, SpawnedLink=SpawnedLink)
+        self.link(receiver=receiver, SpawnedLink=SpawnedLink)
 
     def _notify_links(self):
-        try:
-            while self._links:
-                link = self._links.pop()
-                try:
-                    link(self)
-                except:
-                    traceback.print_exc()
-                    try:
-                        sys.stderr.write('Failed to notify link %s of %r\n\n' % (getfuncname(link), self))
-                    except:
-                        traceback.print_exc()
-        finally:
-            self._notifier = None
+        while self._links:
+            link = self._links.pop()
+            try:
+                link(self)
+            except:
+                self.parent.handle_error((link, self), *sys.exc_info())
+
+
+class _dummy_event(object):
+
+    def stop(self):
+        pass
+
+
+_dummy_event = _dummy_event()
 
 
 def _kill(greenlet, exception, waiter):
     try:
         greenlet.throw(exception)
     except:
-        traceback.print_exc()
+        greenlet.parent.handle_error(greenlet, *sys.exc_info())
     waiter.switch()
 
 
-def joinall(greenlets, timeout=None, raise_error=False):
+def joinall(greenlets, timeout=None, raise_error=False, count=None):
+    xrange = moves.xrange
     from gevent.queue import Queue
     queue = Queue()
     put = queue.put
+    if count is None:
+        count = len(greenlets)
     timeout = Timeout.start_new(timeout)
     try:
         try:
             for greenlet in greenlets:
                 greenlet.rawlink(put)
             if raise_error:
-                for _ in xrange(len(greenlets)):
+                for _ in xrange(count):
                     greenlet = queue.get()
                     if not greenlet.successful():
                         raise greenlet.exception
             else:
-                for _ in xrange(len(greenlets)):
+                for _ in xrange(count):
                     queue.get()
         except:
-            for greenlet in greenlets:
-                greenlet.unlink(put)
             if sys.exc_info()[1] is not timeout:
                 raise
+        finally:
+            for greenlet in greenlets:
+                greenlet.unlink(put)
     finally:
         timeout.cancel()
 
@@ -528,7 +431,7 @@ def _killall3(greenlets, exception, waiter):
             try:
                 g.throw(exception)
             except:
-                traceback.print_exc()
+                g.parent.handle_error(g, *sys.exc_info())
             if not g.dead:
                 diehards.append(g)
     waiter.switch(diehards)
@@ -540,69 +443,29 @@ def _killall(greenlets, exception):
             try:
                 g.throw(exception)
             except:
-                traceback.print_exc()
+                g.parent.handle_error(g, *sys.exc_info())
 
 
 def killall(greenlets, exception=GreenletExit, block=True, timeout=None):
+    if not greenlets:
+        return
+    loop = greenlets[0].loop
     if block:
         waiter = Waiter()
-        core.active_event(_killall3, greenlets, exception, waiter)
-        if block:
-            t = Timeout.start_new(timeout)
-            try:
-                alive = waiter.get()
-                if alive:
-                    joinall(alive, raise_error=False)
-            finally:
-                t.cancel()
+        loop.run_callback(_killall3, greenlets, exception, waiter)
+        t = Timeout.start_new(timeout)
+        try:
+            alive = waiter.get()
+            if alive:
+                joinall(alive, raise_error=False)
+        finally:
+            t.cancel()
     else:
-        core.active_event(_killall, greenlets, exception)
-
-
-class LinkedExited(Exception):
-    pass
-
-
-class LinkedCompleted(LinkedExited):
-    """Raised when a linked greenlet finishes the execution cleanly"""
-
-    msg = "%r completed successfully"
-
-    def __init__(self, source):
-        assert source.ready(), source
-        assert source.successful(), source
-        LinkedExited.__init__(self, self.msg % source)
-
-
-class LinkedKilled(LinkedCompleted):
-    """Raised when a linked greenlet returns GreenletExit instance"""
-
-    msg = "%r returned %s"
-
-    def __init__(self, source):
-        try:
-            result = source.value.__class__.__name__
-        except:
-            result = str(source) or repr(source)
-        LinkedExited.__init__(self, self.msg % (source, result))
-
-
-class LinkedFailed(LinkedExited):
-    """Raised when a linked greenlet dies because of unhandled exception"""
-
-    msg = "%r failed with %s"
-
-    def __init__(self, source):
-        exception = source.exception
-        try:
-            excname = exception.__class__.__name__
-        except:
-            excname = str(exception) or repr(exception)
-        LinkedExited.__init__(self, self.msg % (source, excname))
+        loop.run_callback(_killall, greenlets, exception)
 
 
 def getfuncname(func):
-    if not hasattr(func, 'im_self'):
+    if not hasattr(func, _meth_self):
         try:
             funcname = func.__name__
         except AttributeError:
