@@ -1,4 +1,4 @@
-# Copyright (c) 2009-2010 Denis Bilenko. See LICENSE for details.
+# Copyright (c) 2009-2012 Denis Bilenko. See LICENSE for details.
 """Make the standard library cooperative.
 
 The functions in this module patch parts of the standard library with compatible cooperative counterparts
@@ -58,8 +58,6 @@ Monkey patches:
 
 import sys
 
-noisy = True
-
 __all__ = ['patch_all',
            'patch_socket',
            'patch_ssl',
@@ -69,6 +67,46 @@ __all__ = ['patch_all',
            'patch_thread']
 
 
+def get_unpatched(name, items):
+    # QQQ would prefer to avoid importing gevent.xxx module just to get __target__
+    # XXX does not work for 'time', 'sleep'
+    source = getattr(__import__('gevent.' + name), name)
+    target = getattr(source, '__target__', name)
+    dest = __import__(target)
+    original = getattr(dest, 'monkey_original', dest)
+    if isinstance(items, basestring):
+        return getattr(original, items)
+    else:
+        results = []
+        for item in items:
+            results.append(getattr(original, item))
+        return results
+
+
+class original(object):
+    pass
+
+
+def patch_module(name, items=None):
+    source = getattr(__import__('gevent.' + name), name)
+    target = getattr(source, '__target__', name)
+    dest = __import__(target)
+    monkey_original = dest.monkey_original = original()
+    count = 0
+    if items is None:
+        items = getattr(source, '__implements__', None)
+        if items is None:
+            raise AttributeError('%r does not have __implements__' % source)
+    for attr in items:
+        olditem = getattr(dest, attr, None)
+        newitem = getattr(source, attr)
+        if olditem is not newitem:
+            setattr(monkey_original, attr, olditem)
+            setattr(dest, attr, newitem)
+            count += 1
+    return count
+
+
 def patch_os():
     """Replace :func:`os.fork` with :func:`gevent.fork`."""
     try:
@@ -91,26 +129,25 @@ def patch_thread(threading=True, _threading_local=True):
     If *threading* is true (the default), also patch ``threading.local``.
     If *_threading_local* is true (the default), also patch ``_threading_local.local``.
     """
-    from gevent import thread as green_thread
-    thread = __import__('thread')
-    if thread.exit is not green_thread.exit:
-        thread.get_ident = green_thread.get_ident
-        thread.start_new_thread = green_thread.start_new_thread
-        thread.LockType = green_thread.LockType
-        thread.allocate_lock = green_thread.allocate_lock
-        thread.exit = green_thread.exit
-        if hasattr(green_thread, 'stack_size'):
-            thread.stack_size = green_thread.stack_size
-        from gevent.local import local
-        thread._local = local
-        if threading:
-            if noisy and 'threading' in sys.modules:
-                sys.stderr.write("gevent.monkey's warning: 'threading' is already imported\n\n")
-            threading = __import__('threading')
-            threading.local = local
-        if _threading_local:
-            _threading_local = __import__('_threading_local')
-            _threading_local.local = local
+    if not patch_module('thread'):
+        return
+    from gevent.local import local
+    if threading:
+        from gevent import thread as green_thread
+        threading = __import__('threading')
+        # QQQ Note, that importing threading results in get_hub() call.
+        # QQQ Would prefer not to import threading at all if it's not used;
+        # QQQ that should be possible with import hooks
+        threading.local = local
+        threading._start_new_thread = green_thread.start_new_thread
+        threading._allocate_lock = green_thread.allocate_lock
+        threading.Lock = green_thread.allocate_lock
+        threading._get_ident = green_thread.get_ident
+        from gevent.hub import sleep
+        threading._sleep = sleep
+    if _threading_local:
+        _threading_local = __import__('_threading_local')
+        _threading_local.local = local
 
 
 def patch_socket(dns=True, aggressive=True):
@@ -119,45 +156,28 @@ def patch_socket(dns=True, aggressive=True):
     If *dns* is true, also patch dns functions in :mod:`socket`.
     """
     from gevent import socket
-    _socket = __import__('socket')
-    _socket.socket = socket.socket
-    _socket.SocketType = socket.SocketType
-    _socket.create_connection = socket.create_connection
-    if hasattr(socket, 'socketpair'):
-        _socket.socketpair = socket.socketpair
-    if hasattr(socket, 'fromfd'):
-        _socket.fromfd = socket.fromfd
-    try:
-        from gevent.socket import ssl, sslerror
-        _socket.ssl = ssl
-        _socket.sslerror = sslerror
-    except ImportError:
-        if aggressive:
-            try:
-                del _socket.ssl
-            except AttributeError:
-                pass
+    # Note: although it seems like it's not strictly necessary to monkey patch 'create_connection',
+    # it's better to do it. If 'create_connection' was not monkey patched, but the rest of socket module
+    # was, create_connection would still use "green" getaddrinfo and "green" socket.
+    # However, because gevent.socket.socket.connect is a Python function, the exception raised by it causes
+    # _socket object to be referenced by the frame, thus causing the next invocation of bind(source_address) to fail.
     if dns:
-        patch_dns()
+        items = socket.__implements__
+    else:
+        items = set(socket.__implements__) - set(socket.__dns__)
+    patch_module('socket', items=items)
+    if aggressive:
+        if 'ssl' not in socket.__implements__:
+            socket.__dict__.pop('ssl', None)
 
 
 def patch_dns():
-    from gevent.socket import gethostbyname, getaddrinfo
-    _socket = __import__('socket')
-    _socket.getaddrinfo = getaddrinfo
-    _socket.gethostbyname = gethostbyname
+    from gevent import socket
+    patch_module('socket', items=socket.__dns__)
 
 
 def patch_ssl():
-    try:
-        _ssl = __import__('ssl')
-    except ImportError:
-        return
-    from gevent.ssl import SSLSocket, wrap_socket, get_server_certificate, sslwrap_simple
-    _ssl.SSLSocket = SSLSocket
-    _ssl.wrap_socket = wrap_socket
-    _ssl.get_server_certificate = get_server_certificate
-    _ssl.sslwrap_simple = sslwrap_simple
+    patch_module('ssl')
 
 
 def patch_select(aggressive=False):
@@ -165,24 +185,16 @@ def patch_select(aggressive=False):
 
     If aggressive is true (the default), also remove other blocking functions the :mod:`select`.
     """
-    from gevent.select import select
-    _select = __import__('select')
-    globals()['_select_select'] = _select.select
-    _select.select = select
+    patch_module('select')
     if aggressive:
-        # since these are blocking and don't work with the libevent's event loop
-        # we're removing them here. This makes some other modules (e.g. asyncore)
-        # non-blocking, as they use select that we provide when none of these are available.
-        _select.__dict__.pop('poll', None)
-        _select.__dict__.pop('epoll', None)
-        _select.__dict__.pop('kqueue', None)
-        _select.__dict__.pop('kevent', None)
-
-
-def patch_httplib():
-    from gevent.httplib import HTTPConnection
-    httplib = __import__('httplib')
-    httplib.HTTPConnection = HTTPConnection
+        select = __import__('select')
+        # since these are blocking we're removing them here. This makes some other
+        # modules (e.g. asyncore)  non-blocking, as they use select that we provide
+        # when none of these are available.
+        select.__dict__.pop('poll', None)
+        select.__dict__.pop('epoll', None)
+        select.__dict__.pop('kqueue', None)
+        select.__dict__.pop('kevent', None)
 
 
 def patch_all(socket=True, dns=True, time=True, select=True, thread=True, os=True, ssl=True, httplib=False, aggressive=True):
@@ -199,9 +211,14 @@ def patch_all(socket=True, dns=True, time=True, select=True, thread=True, os=Tru
     if select:
         patch_select(aggressive=aggressive)
     if ssl:
-        patch_ssl()
+        try:
+            patch_ssl()
+        except ImportError:
+            if sys.version_info[:2] > (2, 5):
+                raise
+            # in Python 2.5, 'ssl' is a standalone package not included in stdlib
     if httplib:
-        patch_httplib()
+        raise ValueError('gevent.httplib is no longer provided, httplib must be False')
 
 
 if __name__ == '__main__':
@@ -234,11 +251,11 @@ MONKEY OPTIONS: --verbose %s""" % ', '.join('--[no-]%s' % m for m in modules)
     if verbose:
         import pprint
         import os
-        print 'gevent.monkey.patch_all(%s)' % ', '.join('%s=%s' % item for item in args.items())
-        print 'sys.version=%s' % (sys.version.strip().replace('\n', ' '), )
-        print 'sys.path=%s' % pprint.pformat(sys.path)
-        print 'sys.modules=%s' % pprint.pformat(sorted(sys.modules.keys()))
-        print 'cwd=%s' % os.getcwd()
+        print ('gevent.monkey.patch_all(%s)' % ', '.join('%s=%s' % item for item in args.items()))
+        print ('sys.version=%s' % (sys.version.strip().replace('\n', ' '), ))
+        print ('sys.path=%s' % pprint.pformat(sys.path))
+        print ('sys.modules=%s' % pprint.pformat(sorted(sys.modules.keys())))
+        print ('cwd=%s' % os.getcwd())
 
     patch_all(**args)
     if argv:
@@ -246,4 +263,4 @@ MONKEY OPTIONS: --verbose %s""" % ', '.join('--[no-]%s' % m for m in modules)
         __package__ = None
         execfile(sys.argv[0])
     else:
-        print script_help
+        print (script_help)
