@@ -1,4 +1,4 @@
-# Copyright (c) 2009-2010 Denis Bilenko. See LICENSE for details.
+# Copyright (c) 2009-2011 Denis Bilenko. See LICENSE for details.
 """Managing greenlets in a group.
 
 The :class:`Group` class in this module abstracts a group of running greenlets.
@@ -9,11 +9,15 @@ concurrency: its :meth:`spawn <Pool.spawn>` method blocks if the number of
 greenlets in the pool has already reached the limit, until there is a free slot.
 """
 
-from gevent.hub import GreenletExit, getcurrent
+import sys
+from bisect import insort_right
+
+from gevent.hub import GreenletExit, getcurrent, kill as _kill
 from gevent.greenlet import joinall, Greenlet
 from gevent.timeout import Timeout
 from gevent.event import Event
 from gevent.coros import Semaphore, DummySemaphore
+from gevent import six
 
 __all__ = ['Group', 'Pool']
 
@@ -30,18 +34,14 @@ class Group(object):
         self.greenlets = set(*args)
         if args:
             for greenlet in args[0]:
-                greenlet.rawlink(self.discard)
+                greenlet.rawlink(self._discard)
         # each item we kill we place in dying, to avoid killing the same greenlet twice
         self.dying = set()
         self._empty_event = Event()
         self._empty_event.set()
 
     def __repr__(self):
-        try:
-            classname = self.__class__.__name__
-        except AttributeError:
-            classname = 'Group'  # XXX check if 2.4 really uses this line
-        return '<%s at %s %s>' % (classname, hex(id(self)), self.greenlets)
+        return '<%s at 0x%x %s>' % (self.__class__.__name__, id(self), self.greenlets)
 
     def __len__(self):
         return len(self.greenlets)
@@ -53,39 +53,37 @@ class Group(object):
         return iter(self.greenlets)
 
     def add(self, greenlet):
-        greenlet.rawlink(self.discard)
+        try:
+            rawlink = greenlet.rawlink
+        except AttributeError:
+            pass  # non-Greenlet greenlet, like MAIN
+        else:
+            rawlink(self._discard)
         self.greenlets.add(greenlet)
         self._empty_event.clear()
 
-    def discard(self, greenlet):
+    def _discard(self, greenlet):
         self.greenlets.discard(greenlet)
         self.dying.discard(greenlet)
         if not self.greenlets:
             self._empty_event.set()
 
+    def discard(self, greenlet):
+        self._discard(greenlet)
+        try:
+            unlink = greenlet.unlink
+        except AttributeError:
+            pass  # non-Greenlet greenlet, like MAIN
+        else:
+            unlink(self._discard)
+
     def start(self, greenlet):
         self.add(greenlet)
         greenlet.start()
 
     def spawn(self, *args, **kwargs):
-        add = self.add
-        greenlet = self.greenlet_class.spawn(*args, **kwargs)
-        add(greenlet)
-        return greenlet
-
-    def spawn_link(self, *args, **kwargs):
-        greenlet = self.spawn(*args, **kwargs)
-        greenlet.link()
-        return greenlet
-
-    def spawn_link_value(self, *args, **kwargs):
-        greenlet = self.spawn(*args, **kwargs)
-        greenlet.link_value()
-        return greenlet
-
-    def spawn_link_exception(self, *args, **kwargs):
-        greenlet = self.spawn(*args, **kwargs)
-        greenlet.link_exception()
+        greenlet = self.greenlet_class(*args, **kwargs)
+        self.start(greenlet)
         return greenlet
 
 #     def close(self):
@@ -109,12 +107,18 @@ class Group(object):
                 while self.greenlets:
                     for greenlet in list(self.greenlets):
                         if greenlet not in self.dying:
-                            greenlet.kill(exception, block=False)
+                            try:
+                                kill = greenlet.kill
+                            except AttributeError:
+                                _kill(greenlet, exception)
+                            else:
+                                kill(exception, block=False)
                             self.dying.add(greenlet)
                     if not block:
                         break
                     joinall(self.greenlets)
-            except Timeout, ex:
+            except Timeout:
+                ex = sys.exc_info()[1]
                 if ex is not timer:
                     raise
         finally:
@@ -163,8 +167,7 @@ class Group(object):
             return greenlet
 
     def map(self, func, iterable):
-        greenlets = [self.spawn(func, item) for item in iterable]
-        return [greenlet.get() for greenlet in greenlets]
+        return list(self.imap(func, iterable))
 
     def map_cb(self, func, iterable, callback=None):
         result = self.map(func, iterable)
@@ -182,19 +185,13 @@ class Group(object):
         return Greenlet.spawn(self.map_cb, func, iterable, callback)
 
     def imap(self, func, iterable):
-        """An equivalent of itertools.imap()
-
-        **TODO**: Fix this.
-        """
-        return iter(self.map(func, iterable))
+        """An equivalent of itertools.imap()"""
+        return IMap.spawn(func, iterable, spawn=self.spawn)
 
     def imap_unordered(self, func, iterable):
         """The same as imap() except that the ordering of the results from the
-        returned iterator should be considered arbitrary.
-
-        **TODO**: Fix this.
-        """
-        return iter(self.map(func, iterable))
+        returned iterator should be considered in arbitrary order."""
+        return IMapUnordered.spawn(func, iterable, spawn=self.spawn)
 
     def full(self):
         return False
@@ -203,14 +200,135 @@ class Group(object):
         pass
 
 
-GreenletSet = Group  # the old name; will be deprecated in the future
+class IMapUnordered(Greenlet):
+
+    def __init__(self, func, iterable, spawn=None):
+        from gevent.queue import Queue
+        Greenlet.__init__(self)
+        if spawn is not None:
+            self.spawn = spawn
+        self.func = func
+        self.iterable = iterable
+        self.queue = Queue()
+        self.count = 0
+        self.rawlink(self._on_finish)
+
+    def __iter__(self):
+        return self
+
+    def next(self):
+        value = self.queue.get()
+        if isinstance(value, Failure):
+            raise value.exc
+        return value
+
+    if six.PY3:
+        __next__ = next
+        del next
+
+    def _run(self):
+        try:
+            func = self.func
+            for item in self.iterable:
+                self.count += 1
+                self.spawn(func, item).rawlink(self._on_result)
+        finally:
+            self.__dict__.pop('spawn', None)
+            self.__dict__.pop('func', None)
+            self.__dict__.pop('iterable', None)
+
+    def _on_result(self, greenlet):
+        self.count -= 1
+        if greenlet.successful():
+            self.queue.put(greenlet.value)
+        if self.ready() and self.count <= 0:
+            self.queue.put(Failure(StopIteration))
+
+    def _on_finish(self, _self):
+        if not self.successful():
+            self.queue.put(Failure(self.exception))
+
+
+class IMap(Greenlet):
+
+    def __init__(self, func, iterable, spawn=None):
+        from gevent.queue import Queue
+        Greenlet.__init__(self)
+        if spawn is not None:
+            self.spawn = spawn
+        self.func = func
+        self.iterable = iterable
+        self.queue = Queue()
+        self.count = 0
+        self.waiting = []  # QQQ maybe deque will work faster there?
+        self.index = 0
+        self.maxindex = -1
+        self.rawlink(self._on_finish)
+
+    def __iter__(self):
+        return self
+
+    def next(self):
+        while True:
+            if self.waiting and self.waiting[0][0] <= self.index:
+                index, value = self.waiting.pop(0)
+            else:
+                index, value = self.queue.get()
+                if index > self.index:
+                    insort_right(self.waiting, (index, value))
+                    continue
+            self.index += 1
+            if isinstance(value, Failure):
+                raise value.exc
+            if value is not _SKIP:
+                return value
+
+    if six.PY3:
+        __next__ = next
+        del next
+
+    def _run(self):
+        try:
+            func = self.func
+            for item in self.iterable:
+                self.count += 1
+                g = self.spawn(func, item)
+                g.rawlink(self._on_result)
+                self.maxindex += 1
+                g.index = self.maxindex
+        finally:
+            self.__dict__.pop('spawn', None)
+            self.__dict__.pop('func', None)
+            self.__dict__.pop('iterable', None)
+
+    def _on_result(self, greenlet):
+        self.count -= 1
+        if greenlet.successful():
+            self.queue.put((greenlet.index, greenlet.value))
+        else:
+            self.queue.put((greenlet.index, _SKIP))
+        if self.ready() and self.count <= 0:
+            self.maxindex += 1
+            self.queue.put((self.maxindex, Failure(StopIteration)))
+
+    def _on_finish(self, _self):
+        if not self.successful():
+            self.maxindex += 1
+            self.queue.put((self.maxindex, Failure(self.exception)))
+
+
+class Failure(object):
+    __slots__ = ['exc']
+
+    def __init__(self, exc):
+        self.exc = exc
 
 
 class Pool(Group):
 
     def __init__(self, size=None, greenlet_class=None):
-        if size is not None and size < 1:
-            raise ValueError('Invalid size for pool (positive integer or None required): %r' % (size, ))
+        if size is not None and size < 0:
+            raise ValueError('size must not be negative: %r' % (size, ))
         Group.__init__(self)
         self.size = size
         if greenlet_class is not None:
@@ -231,57 +349,16 @@ class Pool(Group):
             return 1
         return max(0, self.size - len(self))
 
-    def start(self, greenlet):
+    def add(self, greenlet):
         self._semaphore.acquire()
         try:
-            self.add(greenlet)
+            Group.add(self, greenlet)
         except:
             self._semaphore.release()
             raise
-        greenlet.start()
 
-    def spawn(self, *args, **kwargs):
-        self._semaphore.acquire()
-        try:
-            greenlet = self.greenlet_class.spawn(*args, **kwargs)
-            self.add(greenlet)
-        except:
-            self._semaphore.release()
-            raise
-        return greenlet
-
-    def spawn_link(self, *args, **kwargs):
-        self._semaphore.acquire()
-        try:
-            greenlet = self.greenlet_class.spawn_link(*args, **kwargs)
-            self.add(greenlet)
-        except:
-            self._semaphore.release()
-            raise
-        return greenlet
-
-    def spawn_link_value(self, *args, **kwargs):
-        self._semaphore.acquire()
-        try:
-            greenlet = self.greenlet_class.spawn_link_value(*args, **kwargs)
-            self.add(greenlet)
-        except:
-            self._semaphore.release()
-            raise
-        return greenlet
-
-    def spawn_link_exception(self, *args, **kwargs):
-        self._semaphore.acquire()
-        try:
-            greenlet = self.greenlet_class.spawn_link_exception(*args, **kwargs)
-            self.add(greenlet)
-        except:
-            self._semaphore.release()
-            raise
-        return greenlet
-
-    def discard(self, greenlet):
-        Group.discard(self, greenlet)
+    def _discard(self, greenlet):
+        Group._discard(self, greenlet)
         self._semaphore.release()
 
 
@@ -315,3 +392,6 @@ class pass_value(object):
     def __getattr__(self, item):
         assert item != 'callback'
         return getattr(self.callback, item)
+
+
+_SKIP = object()
