@@ -1,5 +1,5 @@
 # Copyright (c) 2005-2009, eventlet contributors
-# Copyright (c) 2009-2010, gevent contributors
+# Copyright (c) 2009-2011, gevent contributors
 
 import errno
 import sys
@@ -41,16 +41,16 @@ def format_date_time(timestamp):
 
 class Input(object):
 
-    def __init__(self, rfile, content_length, wfile=None, chunked_input=False):
+    def __init__(self, rfile, content_length, socket=None, chunked_input=False):
         self.rfile = rfile
         self.content_length = content_length
-        self.wfile = wfile
+        self.socket = socket
         self.position = 0
         self.chunked_input = chunked_input
         self.chunk_length = -1
 
     def _discard(self):
-        if self.wfile is None and (self.position < (self.content_length or 0) or self.chunked_input):
+        if self.socket is None and (self.position < (self.content_length or 0) or self.chunked_input):
             # ## Read and discard body
             while 1:
                 d = self.read(16384)
@@ -58,23 +58,38 @@ class Input(object):
                     break
 
     def _send_100_continue(self):
-        if self.wfile is not None:
-            self.wfile.write(_CONTINUE_RESPONSE)
-            self.wfile = None
+        if self.socket is not None:
+            self.socket.sendall(_CONTINUE_RESPONSE)
+            self.socket = None
 
-    def _do_read(self, reader, length=None):
+    def _do_read(self, length=None, use_readline=False):
+        if use_readline:
+            reader = self.rfile.readline
+        else:
+            reader = self.rfile.read
+        content_length = self.content_length
+        if content_length is None:
+            # Either Content-Length or "Transfer-Encoding: chunked" must be present in a request with a body
+            # if it was chunked, then this function would have not been called
+            return ''
         self._send_100_continue()
-        if length is None and self.content_length is not None:
-            length = self.content_length - self.position
-        if length and length > self.content_length - self.position:
-            length = self.content_length - self.position
+        left = content_length - self.position
+        if length is None:
+            length = left
+        elif length > left:
+            length = left
         if not length:
             return ''
         read = reader(length)
         self.position += len(read)
+        if len(read) < length:
+            if (use_readline and not read.endswith("\n")) or not use_readline:
+                raise IOError("unexpected end of file while reading request at position %s" % (self.position,))
+
         return read
 
-    def _chunked_read(self, rfile, length=None, use_readline=False):
+    def _chunked_read(self, length=None, use_readline=False):
+        rfile = self.rfile
         self._send_100_continue()
 
         if length == 0:
@@ -114,23 +129,26 @@ class Input(object):
                 if use_readline and data[-1] == "\n":
                     break
             else:
-                self.chunk_length = int(rfile.readline().split(";", 1)[0], 16)
+                line = rfile.readline()
+                if not line.endswith("\n"):
+                    self.chunk_length = 0
+                    raise IOError("unexpected end of file while reading chunked data header")
+                self.chunk_length = int(line.split(";", 1)[0], 16)
                 self.position = 0
                 if self.chunk_length == 0:
                     rfile.readline()
-
         return ''.join(response)
 
     def read(self, length=None):
         if self.chunked_input:
-            return self._chunked_read(self.rfile, length)
-        return self._do_read(self.rfile.read, length)
+            return self._chunked_read(length)
+        return self._do_read(length)
 
     def readline(self, size=None):
         if self.chunked_input:
-            return self._chunked_read(self.rfile, size, True)
+            return self._chunked_read(size, True)
         else:
-            return self._do_read(self.rfile.readline, size)
+            return self._do_read(size, use_readline=True)
 
     def readlines(self, hint=None):
         return list(self)
@@ -149,12 +167,14 @@ class WSGIHandler(object):
     protocol_version = 'HTTP/1.1'
     MessageClass = mimetools.Message
 
-    def __init__(self, socket, address, server):
+    def __init__(self, socket, address, server, rfile=None):
         self.socket = socket
         self.client_address = address
         self.server = server
-        self.rfile = socket.makefile('rb', -1)
-        self.wfile = socket.makefile('wb', 0)
+        if rfile is None:
+            self.rfile = socket.makefile('rb', -1)
+        else:
+            self.rfile = rfile
 
     def handle(self):
         try:
@@ -167,7 +187,7 @@ class WSGIHandler(object):
                 if result is True:
                     continue
                 self.status, response_body = result
-                self.wfile.write(response_body)
+                self.socket.sendall(response_body)
                 if self.time_finish == 0:
                     self.time_finish = time.time()
                 self.log_request()
@@ -181,7 +201,6 @@ class WSGIHandler(object):
                     pass
             self.__dict__.pop('socket', None)
             self.__dict__.pop('rfile', None)
-            self.__dict__.pop('wfile', None)
 
     def _check_http_version(self):
         version = self.request_version
@@ -222,13 +241,13 @@ class WSGIHandler(object):
             except KeyError:
                 pass
 
-        content_length = self.headers.get("Content-Length")
+        content_length = self.headers.get("content-length")
         if content_length is not None:
             content_length = int(content_length)
             if content_length < 0:
                 self.log_error('Invalid Content-Length: %r', content_length)
                 return
-            if content_length and self.command in ('GET', 'HEAD'):
+            if content_length and self.command in ('HEAD', ):
                 self.log_error('Unexpected Content-Length')
                 return
 
@@ -267,7 +286,12 @@ class WSGIHandler(object):
         if self.rfile.closed:
             return
 
-        raw_requestline = self.read_requestline()
+        try:
+            raw_requestline = self.read_requestline()
+        except socket.error:
+            # "Connection reset by peer" or other socket errors aren't interesting here
+            return
+
         if not raw_requestline:
             return
 
@@ -279,11 +303,10 @@ class WSGIHandler(object):
         try:
             if not self.read_request(raw_requestline):
                 return ('400', _BAD_REQUEST_RESPONSE)
-        except ValueError, ex:
-            self.log_error('Invalid request: %s', str(ex) or ex.__class__.__name__)
-            return ('400', _BAD_REQUEST_RESPONSE)
-        except Exception, ex:
-            traceback.print_exc()
+        except Exception:
+            ex = sys.exc_info()[1]
+            if not isinstance(ex, ValueError):
+                traceback.print_exc()
             self.log_error('Invalid request: %s', str(ex) or ex.__class__.__name__)
             return ('400', _BAD_REQUEST_RESPONSE)
 
@@ -291,10 +314,12 @@ class WSGIHandler(object):
         self.application = self.server.application
         try:
             self.handle_one_response()
-        except socket.error, ex:
+        except socket.error:
+            ex = sys.exc_info()[1]
             # Broken pipe, connection reset by peer
-            if ex[0] in (errno.EPIPE, errno.ECONNRESET):
+            if ex.args[0] in (errno.EPIPE, errno.ECONNRESET):
                 sys.exc_clear()
+                return
             else:
                 raise
 
@@ -306,50 +331,89 @@ class WSGIHandler(object):
 
         return True  # read more requests
 
+    def finalize_headers(self):
+        if self.provided_date is None:
+            self.response_headers.append(('Date', format_date_time(time.time())))
+
+        if self.code not in (304, 204):
+            # the reply will include message-body; make sure we have either Content-Length or chunked
+            if self.provided_content_length is None:
+                if hasattr(self.result, '__len__'):
+                    self.response_headers.append(('Content-Length', str(sum(len(chunk) for chunk in self.result))))
+                else:
+                    if self.request_version != 'HTTP/1.0':
+                        self.response_use_chunked = True
+                        self.response_headers.append(('Transfer-Encoding', 'chunked'))
+
+    def _sendall(self, data):
+        try:
+            self.socket.sendall(data)
+        except socket.error, ex:
+            self.status = 'socket error: %s' % ex
+            if self.code > 0:
+                self.code = -self.code
+            raise
+        self.response_length += len(data)
+
+    def _write(self, data):
+        if not data:
+            return
+        if self.response_use_chunked:
+            ## Write the chunked encoding
+            data = "%x\r\n%s\r\n" % (len(data), data)
+        self._sendall(data)
+
     def write(self, data):
-        towrite = []
-        if not self.status:
-            raise AssertionError("The application did not call start_response()")
-        if not self.headers_sent:
-            if 'Date' not in self.response_headers_list:
-                self.response_headers.append(('Date', format_date_time(time.time())))
-                self.response_headers_list.append('Date')
+        if self.code in (304, 204) and data:
+            raise AssertionError('The %s response must have no body' % self.code)
 
-            if self.request_version == 'HTTP/1.0' and 'Connection' not in self.response_headers_list:
-                self.response_headers.append(('Connection', 'close'))
-                self.response_headers_list.append('Connection')
-                self.close_connection = True
-            elif ('Connection', 'close') in self.response_headers:
-                self.close_connection = True
+        if self.headers_sent:
+            self._write(data)
+        else:
+            if not self.status:
+                raise AssertionError("The application did not call start_response()")
+            self._write_with_headers(data)
 
-            if self.code not in [204, 304]:
-                # the reply will include message-body; make sure we have either Content-Length or chunked
-                if 'Content-Length' not in self.response_headers_list:
-                    if hasattr(self.result, '__len__'):
-                        self.response_headers.append(('Content-Length', str(sum(len(chunk) for chunk in self.result))))
-                        self.response_headers_list.append('Content-Length')
-                    else:
-                        if self.request_version != 'HTTP/1.0':
-                            self.response_use_chunked = True
-                            self.response_headers.append(('Transfer-Encoding', 'chunked'))
-                            self.response_headers_list.append('Transfer-Encoding')
+    if sys.version_info[:2] >= (2, 6):
+
+        def _write_with_headers(self, data):
+            towrite = bytearray()
+            self.headers_sent = True
+            self.finalize_headers()
+
+            towrite.extend('%s %s\r\n' % (self.request_version, self.status))
+            for header in self.response_headers:
+                towrite.extend('%s: %s\r\n' % header)
+
+            towrite.extend('\r\n')
+            if data:
+                if self.response_use_chunked:
+                    ## Write the chunked encoding
+                    towrite.extend("%x\r\n%s\r\n" % (len(data), data))
+                else:
+                    towrite.extend(data)
+            self._sendall(towrite)
+
+    else:
+        # Python 2.5 does not have bytearray
+
+        def _write_with_headers(self, data):
+            towrite = []
+            self.headers_sent = True
+            self.finalize_headers()
 
             towrite.append('%s %s\r\n' % (self.request_version, self.status))
             for header in self.response_headers:
                 towrite.append('%s: %s\r\n' % header)
 
             towrite.append('\r\n')
-            self.headers_sent = True
-
-        if data:
-            if self.response_use_chunked:
-                ## Write the chunked encoding
-                towrite.append("%x\r\n%s\r\n" % (len(data), data))
-            else:
-                towrite.append(data)
-
-        self.wfile.writelines(towrite)
-        self.response_length += sum(len(x) for x in towrite)
+            if data:
+                if self.response_use_chunked:
+                    ## Write the chunked encoding
+                    towrite.append("%x\r\n%s\r\n" % (len(data), data))
+                else:
+                    towrite.append(data)
+            self._sendall(''.join(towrite))
 
     def start_response(self, status, headers, exc_info=None):
         if exc_info:
@@ -362,8 +426,32 @@ class WSGIHandler(object):
                 exc_info = None
         self.code = int(status.split(' ', 1)[0])
         self.status = status
-        self.response_headers = [('-'.join([x.capitalize() for x in key.split('-')]), value) for key, value in headers]
-        self.response_headers_list = [x[0] for x in self.response_headers]
+        self.response_headers = headers
+
+        provided_connection = None
+        self.provided_date = None
+        self.provided_content_length = None
+
+        for header, value in headers:
+            header = header.lower()
+            if header == 'connection':
+                provided_connection = value
+            elif header == 'date':
+                self.provided_date = value
+            elif header == 'content-length':
+                self.provided_content_length = value
+
+        if self.request_version == 'HTTP/1.0' and provided_connection is None:
+            headers.append(('Connection', 'close'))
+            self.close_connection = True
+        elif provided_connection == 'close':
+            self.close_connection = True
+
+        if self.code in (304, 204):
+            if self.provided_content_length is not None and self.provided_content_length != '0':
+                msg = 'Invalid Content-Length for %s response: %r (must be absent or zero)' % (self.code, self.provided_content_length)
+                raise AssertionError(msg)
+
         return self.write
 
     def log_request(self):
@@ -373,18 +461,16 @@ class WSGIHandler(object):
 
     def format_request(self):
         now = datetime.now().replace(microsecond=0)
+        length = self.response_length or '-'
         if self.time_finish:
             delta = '%.6f' % (self.time_finish - self.time_start)
-            length = self.response_length
         else:
             delta = '-'
-            if not self.response_length:
-                length = '-'
         return '%s - - [%s] "%s" %s %s %s' % (
             self.client_address[0],
             now,
             self.requestline,
-            (self.status or '000').split()[0],
+            (getattr(self, 'status', None) or '000').split()[0],
             length,
             delta)
 
@@ -395,7 +481,7 @@ class WSGIHandler(object):
         if self.status and not self.headers_sent:
             self.write('')
         if self.response_use_chunked:
-            self.wfile.writelines('0\r\n\r\n')
+            self.socket.sendall('0\r\n\r\n')
             self.response_length += 5
 
     def run_application(self):
@@ -414,30 +500,44 @@ class WSGIHandler(object):
         try:
             try:
                 self.run_application()
-            except GreenletExit:
-                raise
-            except Exception:
-                traceback.print_exc()
-                sys.exc_clear()
-                try:
-                    args = (getattr(self, 'server', ''),
-                            getattr(self, 'requestline', ''),
-                            getattr(self, 'client_address', ''),
-                            getattr(self, 'application', ''))
-                    msg = '%s: Failed to handle request:\n  request = %s from %s\n  application = %s\n\n' % args
-                    sys.stderr.write(msg)
-                except Exception:
-                    sys.exc_clear()
-                if not self.response_length:
-                    self.start_response(_INTERNAL_ERROR_STATUS, _INTERNAL_ERROR_HEADERS)
-                    self.write(_INTERNAL_ERROR_BODY)
+            finally:
+                close = getattr(self.result, 'close', None)
+                if close is not None:
+                    close()
+                self.wsgi_input._discard()
+        except:
+            self.handle_error(*sys.exc_info())
         finally:
-            if hasattr(self.result, 'close'):
-                self.result.close()
-            self.wsgi_input._discard()
             self.time_finish = time.time()
             self.log_request()
 
+    def handle_error(self, type, value, tb):
+        if not issubclass(type, GreenletExit):
+            self.server.loop.handle_error(self.environ, type, value, tb)
+        del tb
+        if self.response_length:
+            self.close_connection = True
+        else:
+            self.start_response(_INTERNAL_ERROR_STATUS, _INTERNAL_ERROR_HEADERS)
+            self.write(_INTERNAL_ERROR_BODY)
+
+    def _headers(self):
+        key = None
+        value = None
+        for header in self.headers.headers:
+            if key is not None and header[:1] in " \t":
+                value += header
+                continue
+
+            if key not in (None, 'CONTENT_TYPE', 'CONTENT_LENGTH'):
+                yield 'HTTP_' + key, value.strip()
+
+            key, value = header.split(':', 1)
+            key = key.replace('-', '_').upper()
+
+        if key not in (None, 'CONTENT_TYPE', 'CONTENT_LENGTH'):
+            yield 'HTTP_' + key, value.strip()
+
     def get_environ(self):
         env = self.server.get_environ()
         env['REQUEST_METHOD'] = self.command
@@ -458,32 +558,26 @@ class WSGIHandler(object):
             env['CONTENT_LENGTH'] = length
         env['SERVER_PROTOCOL'] = 'HTTP/1.0'
 
-        host, port = self.socket.getsockname()
-        env['SERVER_NAME'] = host
-        env['SERVER_PORT'] = str(port)
-        env['REMOTE_ADDR'] = self.client_address[0]
-        env['GATEWAY_INTERFACE'] = 'CGI/1.1'
+        client_address = self.client_address
+        if isinstance(client_address, tuple):
+            env['REMOTE_ADDR'] = str(client_address[0])
+            env['REMOTE_PORT'] = str(client_address[1])
 
-        for header in self.headers.headers:
-            key, value = header.split(':', 1)
-            key = key.replace('-', '_').upper()
-            if key not in ('CONTENT_TYPE', 'CONTENT_LENGTH'):
-                value = value.strip()
-                key = 'HTTP_' + key
-                if key in env:
-                    if 'COOKIE' in key:
-                        env[key] += '; ' + value
-                    else:
-                        env[key] += ',' + value
+        for key, value in self._headers():
+            if key in env:
+                if 'COOKIE' in key:
+                    env[key] += '; ' + value
                 else:
-                    env[key] = value
+                    env[key] += ',' + value
+            else:
+                env[key] = value
 
         if env.get('HTTP_EXPECT') == '100-continue':
-            wfile = self.wfile
+            socket = self.socket
         else:
-            wfile = None
+            socket = None
         chunked = env.get('HTTP_TRANSFER_ENCODING', '').lower() == 'chunked'
-        self.wsgi_input = Input(self.rfile, self.content_length, wfile=wfile, chunked_input=chunked)
+        self.wsgi_input = Input(self.rfile, self.content_length, socket=socket, chunked_input=chunked)
         env['wsgi.input'] = self.wsgi_input
         return env
 
@@ -512,6 +606,7 @@ class WSGIServer(StreamServer):
         else:
             self.log = log
         self.set_environ(environ)
+        self.set_max_accept()
 
     def set_environ(self, environ=None):
         if environ is not None:
@@ -527,14 +622,30 @@ class WSGIServer(StreamServer):
         if self.environ.get('wsgi.errors') is None:
             self.environ['wsgi.errors'] = sys.stderr
 
+    def set_max_accept(self):
+        if self.environ.get('wsgi.multiprocess'):
+            self.max_accept = 1
+
     def get_environ(self):
         return self.environ.copy()
 
-    def pre_start(self):
-        StreamServer.pre_start(self)
-        if 'SERVER_NAME' not in self.environ:
-            self.environ['SERVER_NAME'] = socket.getfqdn(self.server_host)
-        self.environ.setdefault('SERVER_PORT', str(self.server_port))
+    def init_socket(self):
+        StreamServer.init_socket(self)
+        self.update_environ()
+
+    def update_environ(self):
+        address = self.address
+        if isinstance(address, tuple):
+            if 'SERVER_NAME' not in self.environ:
+                try:
+                    name = socket.getfqdn(address[0])
+                except socket.error:
+                    name = str(address[0])
+                self.environ['SERVER_NAME'] = name
+            self.environ.setdefault('SERVER_PORT', str(address[1]))
+        else:
+            self.environ.setdefault('SERVER_NAME', '')
+            self.environ.setdefault('SERVER_PORT', '')
 
     def handle(self, socket, address):
         handler = self.handler_class(socket, address, self)
