@@ -26,9 +26,10 @@ import re
 from gevent import sleep, with_timeout, getcurrent
 from gevent import greenlet
 from gevent.event import AsyncResult
-from gevent.queue import Queue
+from gevent.queue import Queue, Channel
 
 DELAY = 0.01
+greentest.TestCase.error_fatal = False
 
 
 class ExpectedError(Exception):
@@ -49,31 +50,13 @@ class TestLink(greentest.TestCase):
             raise
         raise AssertionError('%s not raised, returned %r' % (err, result))
 
-    def test_link_to_greenlet(self):
-        p = gevent.spawn(lambda: 100)
-        receiver = gevent.spawn(sleep, 1)
-        p.link(receiver)
-        self.assertRaises(greenlet.LinkedCompleted, receiver.get)
-        receiver2 = gevent.spawn(sleep, 1)
-        p.link(receiver2)
-        self.assertRaises(greenlet.LinkedCompleted, receiver2.get)
-
-    def test_link_to_inactive_greenlet(self):
-        p = gevent.spawn(lambda: 100)
-        receiver = gevent.spawn_later(10000, sleep, 1)
-        try:
-            p.link(receiver)
-            self.assertRaises(greenlet.LinkedCompleted, receiver.get)
-        finally:
-            receiver.kill()
-
     def test_link_to_asyncresult(self):
         p = gevent.spawn(lambda: 100)
         event = AsyncResult()
         p.link(event)
         self.assertEqual(event.get(), 100)
 
-        for i in xrange(3):
+        for i in range(3):
             event2 = AsyncResult()
             p.link(event2)
             self.assertEqual(event2.get(), 100)
@@ -85,10 +68,10 @@ class TestLink(greentest.TestCase):
         p.link(event)
         self.assertRaises(err, event.get)
 
-        for i in xrange(3):
+        for i in range(3):
             event2 = AsyncResult()
             p.link(event2)
-            self.assertRaises(err, event.get)
+            self.assertRaises(err, event2.get)
 
     def test_link_to_queue(self):
         p = gevent.spawn(lambda: 100)
@@ -96,7 +79,7 @@ class TestLink(greentest.TestCase):
         p.link(q.put)
         self.assertEqual(q.get().get(), 100)
 
-        for i in xrange(3):
+        for i in range(3):
             p.link(q.put)
             self.assertEqual(q.get().get(), 100)
 
@@ -104,32 +87,26 @@ class TestLink(greentest.TestCase):
         p1 = gevent.spawn(lambda: 101)
         p2 = gevent.spawn(lambda: 102)
         p3 = gevent.spawn(lambda: 103)
-        q = Queue(0)
+        q = Channel()
         p1.link(q.put)
         p2.link(q.put)
         p3.link(q.put)
         results = [q.get().get(), q.get().get(), q.get().get()]
         assert sorted(results) == [101, 102, 103], results
 
-    def test_link_to_current(self):
-        p = gevent.spawn(lambda: 100)
-        p.link()
-        self.assertRaises(greenlet.LinkedCompleted, sleep, 0.1)
-        self.assertRaises(greenlet.LinkedCompleted, p.link)
-
 
 class TestUnlink(greentest.TestCase):
     switch_expected = False
 
     def setUp(self):
         greentest.TestCase.setUp(self)
-        self.p = gevent.spawn(test_func)
+        self.p = gevent.spawn(dummy_test_func)
 
     def _test_func(self, link):
         p = self.p
-        link(test_func)
+        link(dummy_test_func)
         assert len(p._links) == 1, p._links
-        p.unlink(test_func)
+        p.unlink(dummy_test_func)
         assert not p._links, p._links
 
         link(self.setUp)
@@ -146,58 +123,19 @@ class TestUnlink(greentest.TestCase):
     def test_func_link_exception(self):
         self._test_func(self.p.link_exception)
 
-    def _test_greenlet(self, link):
-        p = self.p
-        link(getcurrent())
-        assert len(p._links) == 1, p._links
-        p.unlink(getcurrent())
-        assert not p._links, p._links
-
-        g = gevent.Greenlet()
-        link(g)
-        assert len(p._links) == 1, p._links
-        p.unlink(g)
-        assert not p._links, p._links
-
-    def test_greenlet_link(self):
-        self._test_greenlet(self.p.link)
-
-    def test_greenlet_link_value(self):
-        self._test_greenlet(self.p.link_value)
-
-    def test_greenlet_link_exception(self):
-        self._test_greenlet(self.p.link_exception)
-
 
 class LinksTestCase(greentest.TestCase):
 
     def link(self, p, listener=None):
         getattr(p, self.link_method)(listener)
 
-    def receiverf(self, proc_flag):
-        sleep(DELAY)
-        proc_flag.append('finished')
-
-    def set_links(self, p, first_time, kill_exc_type):
+    def set_links(self, p):
         event = AsyncResult()
         self.link(p, event)
 
-        proc_flag = []
-
-        receiver = gevent.spawn(self.receiverf, proc_flag)
-        self.link(p, receiver)
-
         queue = Queue(1)
         self.link(p, queue.put)
 
-        try:
-            self.link(p)
-        except kill_exc_type:
-            if first_time:
-                raise
-        else:
-            assert first_time, 'not raising here only first time'
-
         callback_flag = ['initial']
         self.link(p, lambda *args: callback_flag.remove('initial'))
 
@@ -205,32 +143,20 @@ class LinksTestCase(greentest.TestCase):
             self.link(p, AsyncResult())
             self.link(p, Queue(1).put)
 
-        return event, receiver, proc_flag, queue, callback_flag
-
-    def myprocf(self, proc_finished_flag):
-        sleep(10)
-        proc_finished_flag.append('finished')
-        return 555
+        return event, queue, callback_flag
 
     def set_links_timeout(self, link):
         # stuff that won't be touched
         event = AsyncResult()
         link(event)
 
-        proc_finished_flag = []
-        myproc = gevent.spawn(self.myprocf, proc_finished_flag)
-        link(myproc)
-
-        queue = Queue(0)
+        queue = Channel()
         link(queue.put)
-        return event, myproc, proc_finished_flag, queue
+        return event, queue
 
-    def check_timed_out(self, event, myproc, proc_finished_flag, queue):
+    def check_timed_out(self, event, queue):
         assert with_timeout(DELAY, event.get, timeout_value=X) is X, repr(event.get())
         assert with_timeout(DELAY, queue.get, timeout_value=X) is X, queue.get()
-        assert with_timeout(DELAY, gevent.joinall, [myproc], timeout_value=X) is X
-        assert proc_finished_flag == [], proc_finished_flag
-        myproc.kill()
 
 
 def return25():
@@ -245,74 +171,53 @@ class TestReturn_link(LinksTestCase):
     link_method = 'link'
 
     def cleanup(self):
-        self.p._links.clear()
+        while self.p._links:
+            self.p._links.pop()
 
     def test_return(self):
         self.p = gevent.spawn(return25)
-        self._test_return(self.p, True, 25, greenlet.LinkedCompleted, sleep0)
-        # repeating the same with dead process
-        for _ in xrange(3):
-            self._test_return(self.p, False, 25, greenlet.LinkedCompleted, sleep0)
-        self.p._links.clear()
+        for _ in range(3):
+            self._test_return(self.p, 25, sleep0)
+        self.cleanup()
         self.p.kill()
 
-    def _test_return(self, p, first_time, result, kill_exc_type, action):
-        event, receiver, proc_flag, queue, callback_flag = self.set_links(p, first_time, kill_exc_type)
+    def _test_return(self, p, result, action):
+        event, queue, callback_flag = self.set_links(p)
 
         # stuff that will time out because there's no unhandled exception:
         xxxxx = self.set_links_timeout(p.link_exception)
 
-        try:
-            sleep(DELAY * 2)
-        except kill_exc_type:
-            assert first_time, 'raising here only first time'
-        else:
-            assert not first_time, 'Should not raise LinkedKilled here after first time'
-
+        sleep(DELAY * 2)
         assert not p, p
 
         self.assertEqual(event.get(), result)
         self.assertEqual(queue.get().get(), result)
-        self.assertRaises(kill_exc_type, receiver.get)
-        self.assertRaises(kill_exc_type, gevent.joinall, [receiver], raise_error=True)
 
         sleep(DELAY)
-        assert not proc_flag, proc_flag
         assert not callback_flag, callback_flag
 
         self.check_timed_out(*xxxxx)
 
-    def _test_kill(self, p, first_time, kill_exc_type):
-        event, receiver, proc_flag, queue, callback_flag = self.set_links(p, first_time, kill_exc_type)
+    def _test_kill(self, p):
+        event, queue, callback_flag = self.set_links(p)
         xxxxx = self.set_links_timeout(p.link_exception)
 
         p.kill()
-        try:
-            sleep(DELAY)
-        except kill_exc_type:
-            assert first_time, 'raising here only first time'
-        else:
-            assert not first_time, 'Should not raise LinkedKilled here after first time'
-
+        sleep(DELAY)
         assert not p, p
 
         assert isinstance(event.get(), greenlet.GreenletExit), event.get()
         assert isinstance(queue.get().get(), greenlet.GreenletExit), queue.get().get()
-        self.assertRaises(kill_exc_type, gevent.joinall, [receiver], raise_error=True)
-        self.assertRaises(kill_exc_type, receiver.get)
 
         sleep(DELAY)
-        assert not proc_flag, proc_flag
         assert not callback_flag, callback_flag
 
         self.check_timed_out(*xxxxx)
 
     def test_kill(self):
         p = self.p = gevent.spawn(sleep, DELAY)
-        self._test_kill(p, True, greenlet.LinkedKilled)
-        # repeating the same with dead process
-        for _ in xrange(3):
-            self._test_kill(p, False, greenlet.LinkedKilled)
+        for _ in range(3):
+            self._test_kill(p)
 
 
 class TestReturn_link_value(TestReturn_link):
@@ -322,35 +227,24 @@ class TestReturn_link_value(TestReturn_link):
 class TestRaise_link(LinksTestCase):
     link_method = 'link'
 
-    def _test_raise(self, p, first_time, kill_exc_type):
-        event, receiver, proc_flag, queue, callback_flag = self.set_links(p, first_time, kill_exc_type)
+    def _test_raise(self, p):
+        event, queue, callback_flag = self.set_links(p)
         xxxxx = self.set_links_timeout(p.link_value)
 
-        try:
-            sleep(DELAY)
-        except kill_exc_type:
-            assert first_time, 'raising here only first time'
-        else:
-            assert not first_time, 'Should not raise LinkedKilled here after first time'
-
+        sleep(DELAY)
         assert not p, p
 
         self.assertRaises(ExpectedError, event.get)
         self.assertEqual(queue.get(), p)
-        self.assertRaises(kill_exc_type, receiver.get)
-        self.assertRaises(kill_exc_type, gevent.joinall, [receiver], raise_error=True)
         sleep(DELAY)
-        assert not proc_flag, proc_flag
         assert not callback_flag, callback_flag
 
         self.check_timed_out(*xxxxx)
 
     def test_raise(self):
         p = self.p = gevent.spawn(lambda: getcurrent().throw(ExpectedError('test_raise')))
-        self._test_raise(p, True, greenlet.LinkedFailed)
-        # repeating the same with dead process
-        for _ in xrange(3):
-            self._test_raise(p, False, greenlet.LinkedFailed)
+        for _ in range(3):
+            self._test_raise(p)
 
 
 class TestRaise_link_exception(TestRaise_link):
@@ -379,15 +273,8 @@ class TestStuff(greentest.TestCase):
             sleep(DELAY)
             return 1
         x = gevent.spawn(x)
-        z = gevent.spawn(lambda: 3)
         y = gevent.spawn(lambda: getcurrent().throw(ExpectedError('test_wait_error')))
-        y.link(x)
-        x.link(y)
-        y.link(z)
-        z.link(y)
-        self.assertRaises(ExpectedError, gevent.joinall, [x, y, z], raise_error=True)
-        self.assertRaises(greenlet.LinkedFailed, gevent.joinall, [x], raise_error=True)
-        self.assertEqual(z.get(), 3)
+        self.assertRaises(ExpectedError, gevent.joinall, [x, y], raise_error=True)
         self.assertRaises(ExpectedError, gevent.joinall, [y], raise_error=True)
 
     def test_joinall_exception_order(self):
@@ -399,7 +286,8 @@ class TestStuff(greentest.TestCase):
         b = gevent.spawn(lambda: getcurrent().throw(ExpectedError('second')))
         try:
             gevent.joinall([a, b], raise_error=True)
-        except ExpectedError, ex:
+        except ExpectedError:
+            ex = sys.exc_info()[1]
             assert 'second' in str(ex), repr(str(ex))
         gevent.joinall([a, b])
 
@@ -487,19 +375,12 @@ class TestStuff(greentest.TestCase):
                 raise ExpectedError('test_killing_unlinked')
             except:
                 e.set_exception(sys.exc_info()[1])
+                gevent.sleep(0)
 
-        p = gevent.spawn_link(func)
-        try:
-            try:
-                e.wait()
-            except ExpectedError:
-                pass
-        finally:
-            p.unlink()  # this disables LinkedCompleted that otherwise would be raised by the next line
         sleep(DELAY)
 
 
-def test_func(*args):
+def dummy_test_func(*args):
     pass
 
 
@@ -514,12 +395,12 @@ hexobj = re.compile('-?0x[0123456789abcdef]+L?', re.I)
 class TestStr(greentest.TestCase):
 
     def test_function(self):
-        g = gevent.Greenlet.spawn(test_func)
-        self.assertEqual(hexobj.sub('X', str(g)), '<Greenlet at X: test_func>')
+        g = gevent.Greenlet.spawn(dummy_test_func)
+        self.assertEqual(hexobj.sub('X', str(g)), '<Greenlet at X: dummy_test_func>')
         assert_not_ready(g)
         g.join()
         assert_ready(g)
-        self.assertEqual(hexobj.sub('X', str(g)), '<Greenlet at X: test_func>')
+        self.assertEqual(hexobj.sub('X', str(g)), '<Greenlet at X: dummy_test_func>')
 
     def test_method(self):
         g = gevent.Greenlet.spawn(A().method)
@@ -575,7 +456,7 @@ class TestBasic(greentest.TestCase):
 
         g = gevent.Greenlet(func, 0.01, return_value=5)
         g.link(lambda x: link_test.append(x))
-        assert not g
+        assert not g, bool(g)
         assert not g.dead
         assert not g.started
         assert not g.ready()
@@ -584,7 +465,7 @@ class TestBasic(greentest.TestCase):
         assert g.exception is None
 
         g.start()
-        assert not g
+        assert g  # changed
         assert not g.dead
         assert g.started  # changed
         assert not g.ready()
@@ -593,7 +474,7 @@ class TestBasic(greentest.TestCase):
         assert g.exception is None
 
         gevent.sleep(0.001)
-        assert g  # changed
+        assert g
         assert not g.dead
         assert g.started
         assert not g.ready()
@@ -718,6 +599,23 @@ class TestBasic(greentest.TestCase):
         self._test_kill_running(block=False)
 
 
+class TestStart(greentest.TestCase):
+
+    def test(self):
+        g = gevent.spawn(gevent.sleep, 0.01)
+        assert g.started
+        assert not g.dead
+        g.start()
+        assert g.started
+        assert not g.dead
+        g.join()
+        assert not g.started
+        assert g.dead
+        g.start()
+        assert not g.started
+        assert g.dead
+
+
 def assert_ready(g):
     assert g.dead, g
     assert g.ready(), g
