@@ -6,6 +6,10 @@ from gevent import pool
 DELAY = 0.1
 
 
+class SpecialError(Exception):
+    pass
+
+
 class Undead(object):
 
     def __init__(self):
@@ -15,6 +19,8 @@ class Undead(object):
         while True:
             try:
                 gevent.sleep(1)
+            except SpecialError:
+                break
             except:
                 self.shot_count += 1
 
@@ -23,7 +29,7 @@ class Test(greentest.TestCase):
 
     def test_basic(self):
         DELAY = 0.05
-        s = pool.GreenletSet()
+        s = pool.Group()
         s.spawn(gevent.sleep, DELAY)
         assert len(s) == 1, s
         s.spawn(gevent.sleep, DELAY * 2.)
@@ -34,7 +40,7 @@ class Test(greentest.TestCase):
         assert not s, s
 
     def test_waitall(self):
-        s = pool.GreenletSet()
+        s = pool.Group()
         s.spawn(gevent.sleep, DELAY)
         s.spawn(gevent.sleep, DELAY * 2)
         assert len(s) == 2, s
@@ -46,19 +52,19 @@ class Test(greentest.TestCase):
         assert DELAY * 1.9 <= delta <= DELAY * 2.5, (delta, DELAY)
 
     def test_kill_block(self):
-        s = pool.GreenletSet()
+        s = pool.Group()
         s.spawn(gevent.sleep, DELAY)
         s.spawn(gevent.sleep, DELAY * 2)
         assert len(s) == 2, s
         start = time.time()
-        s.kill(block=True)
+        s.kill()
         assert not s, s
         assert len(s) == 0, s
         delta = time.time() - start
         assert delta < DELAY * 0.8, delta
 
     def test_kill_noblock(self):
-        s = pool.GreenletSet()
+        s = pool.Group()
         s.spawn(gevent.sleep, DELAY)
         s.spawn(gevent.sleep, DELAY * 2)
         assert len(s) == 2, s
@@ -83,7 +89,7 @@ class Test(greentest.TestCase):
             self.assertEqual(u2.shot_count, count2)
 
         gevent.sleep(0.01)
-        s = pool.GreenletSet([p1, p2])
+        s = pool.Group([p1, p2])
         assert len(s) == 2, s
         check(0, 0)
         s.killone(p1, block=False)
@@ -106,11 +112,14 @@ class Test(greentest.TestCase):
         assert len(s) == 2, s
         check(1, 1)
 
+        p1.kill(SpecialError)
+        p2.kill(SpecialError)
+
     def test_killall_subclass(self):
         p1 = GreenletSubclass.spawn(lambda: 1 / 0)
         p2 = GreenletSubclass.spawn(lambda: gevent.sleep(10))
-        s = pool.GreenletSet([p1, p2])
-        s.kill(block=True)
+        s = pool.Group([p1, p2])
+        s.kill()
 
 
 class GreenletSubclass(gevent.Greenlet):
