@@ -1,11 +1,17 @@
+from __future__ import with_statement
 from time import time
 import gevent
 from gevent import pool
 from gevent.event import Event
 import greentest
+import random
+from greentest import ExpectedException
+import six
+
+import unittest
 
 
-class TestCoroutinePool(greentest.TestCase):
+class TestCoroutinePool(unittest.TestCase):
     klass = pool.Pool
 
     def test_apply_async(self):
@@ -33,6 +39,7 @@ class TestCoroutinePool(greentest.TestCase):
         results = []
 
         def producer():
+            gevent.sleep(0.001)
             results.append('prod')
             evt.set()
 
@@ -45,7 +52,7 @@ class TestCoroutinePool(greentest.TestCase):
         done = pool.spawn(consumer)
         pool.apply_async(producer)
         done.get()
-        self.assertEquals(['cons1', 'prod', 'cons2'], results)
+        self.assertEqual(['cons1', 'prod', 'cons2'], results)
 
     def dont_test_timer_cancel(self):
         timer_fired = []
@@ -59,24 +66,14 @@ class TestCoroutinePool(greentest.TestCase):
         pool = self.klass(2)
         pool.apply(some_work)
         gevent.sleep(0)
-        self.assertEquals(timer_fired, [])
+        self.assertEqual(timer_fired, [])
 
     def test_reentrant(self):
         pool = self.klass(1)
-
-        def reenter():
-            result = pool.apply(lambda a: a, ('reenter', ))
-            self.assertEqual('reenter', result)
-
-        pool.apply(reenter)
-
+        result = pool.apply(pool.apply, (lambda a: a + 1, (5, )))
+        self.assertEqual(result, 6)
         evt = Event()
-
-        def reenter_async():
-            pool.apply_async(lambda a: a, ('reenter', ))
-            evt.set()
-
-        pool.apply_async(reenter_async)
+        pool.apply_async(evt.set)
         evt.wait()
 
     def test_stderr_raising(self):
@@ -86,19 +83,14 @@ class TestCoroutinePool(greentest.TestCase):
         import sys
         pool = self.klass(size=1)
 
-        def crash(*args, **kw):
-            raise RuntimeError("Whoa")
-
-        class FakeFile(object):
-            write = crash
-
         # we're going to do this by causing the traceback.print_exc in
         # safe_apply to raise an exception and thus exit _main_loop
         normal_err = sys.stderr
         try:
             sys.stderr = FakeFile()
             waiter = pool.spawn(crash)
-            self.assertRaises(RuntimeError, waiter.get)
+            with gevent.Timeout(2):
+                self.assertRaises(RuntimeError, waiter.get)
             # the pool should have something free at this point since the
             # waiter returned
             # pool.Pool change: if an exception is raised during execution of a link,
@@ -118,6 +110,16 @@ class TestCoroutinePool(greentest.TestCase):
             pool.join()
 
 
+def crash(*args, **kw):
+    raise RuntimeError("Whoa")
+
+
+class FakeFile(object):
+
+    def write(*args):
+        raise RuntimeError('Whaaa')
+
+
 class PoolBasicTests(greentest.TestCase):
     klass = pool.Pool
 
@@ -146,16 +148,51 @@ class PoolBasicTests(greentest.TestCase):
         p.apply_async(r.append, (4, ))
         self.assertEqual(r, [1])
         gevent.sleep(0.01)
-        self.assertEqual(r, [1, 2, 3, 4])
+        self.assertEqual(sorted(r), [1, 2, 3, 4])
 
-    def test_execute(self):
+    def test_discard(self):
+        p = self.klass(size=1)
+        first = p.spawn(gevent.sleep, 1000)
+        p.discard(first)
+        first.kill()
+        assert not first, first
+        self.assertEqual(len(p), 0)
+        self.assertEqual(p._semaphore.counter, 1)
+
+    def test_add_method(self):
+        p = self.klass(size=1)
+        first = gevent.spawn(gevent.sleep, 1000)
+        try:
+            second = gevent.spawn(gevent.sleep, 1000)
+            try:
+                self.assertEqual(p.free_count(), 1)
+                self.assertEqual(len(p), 0)
+                p.add(first)
+                timeout = gevent.Timeout(0.1)
+                timeout.start()
+                try:
+                    p.add(second)
+                except gevent.Timeout:
+                    pass
+                else:
+                    raise AssertionError('Expected timeout')
+                finally:
+                    timeout.cancel()
+                self.assertEqual(p.free_count(), 0)
+                self.assertEqual(len(p), 1)
+            finally:
+                second.kill()
+        finally:
+            first.kill()
+
+    def test_apply(self):
         p = self.klass()
         result = p.apply(lambda a: ('foo', a), (1, ))
         self.assertEqual(result, ('foo', 1))
 
-    def test_init_zerosize(self):
+    def test_init_error(self):
         self.switch_expected = False
-        self.assertRaises(ValueError, self.klass, 0)
+        self.assertRaises(ValueError, self.klass, -1)
 
 #
 # tests from standard library test/test_multiprocessing.py
@@ -179,27 +216,39 @@ def sqr(x, wait=0.0):
     gevent.sleep(wait)
     return x * x
 
+
+def squared(x):
+    return x * x
+
+
+def sqr_random_sleep(x):
+    gevent.sleep(random.random() * 0.1)
+    return x * x
+
+
 TIMEOUT1, TIMEOUT2, TIMEOUT3 = 0.082, 0.035, 0.14
 
 
 class TestPool(greentest.TestCase):
+    __timeout__ = 5
     size = 1
 
     def setUp(self):
+        greentest.TestCase.setUp(self)
         self.pool = pool.Pool(self.size)
 
-    def tearDown(self):
+    def cleanup(self):
         self.pool.join()
 
     def test_apply(self):
         papply = self.pool.apply
-        self.assertEqual(papply(sqr, (5,)), sqr(5))
-        self.assertEqual(papply(sqr, (), {'x': 3}), sqr(x=3))
+        self.assertEqual(papply(sqr, (5,)), 25)
+        self.assertEqual(papply(sqr, (), {'x': 3}), 9)
 
     def test_map(self):
         pmap = self.pool.map
-        self.assertEqual(pmap(sqr, range(10)), map(sqr, range(10)))
-        self.assertEqual(pmap(sqr, range(100)), map(sqr, range(100)))
+        self.assertEqual(pmap(sqr, range(10)), list(map(squared, range(10))))
+        self.assertEqual(pmap(sqr, range(100)), list(map(squared, range(100))))
 
     def test_async(self):
         res = self.pool.apply_async(sqr, (7, TIMEOUT1,))
@@ -221,33 +270,66 @@ class TestPool(greentest.TestCase):
         get = TimingWrapper(res.get)
         self.assertRaises(gevent.Timeout, get, timeout=TIMEOUT2)
         self.assertAlmostEqual(get.elapsed, TIMEOUT2, 1)
+        self.pool.join()
 
     def test_imap(self):
         it = self.pool.imap(sqr, range(10))
-        self.assertEqual(list(it), map(sqr, range(10)))
+        self.assertEqual(list(it), list(map(squared, range(10))))
 
         it = self.pool.imap(sqr, range(10))
         for i in range(10):
-            self.assertEqual(it.next(), i * i)
-        self.assertRaises(StopIteration, it.next)
+            self.assertEqual(six.advance_iterator(it), i * i)
+        self.assertRaises(StopIteration, lambda: six.advance_iterator(it))
 
         it = self.pool.imap(sqr, range(1000))
         for i in range(1000):
-            self.assertEqual(it.next(), i * i)
-        self.assertRaises(StopIteration, it.next)
+            self.assertEqual(six.advance_iterator(it), i * i)
+        self.assertRaises(StopIteration, lambda: six.advance_iterator(it))
+
+    def test_imap_random(self):
+        it = self.pool.imap(sqr_random_sleep, range(10))
+        self.assertEqual(list(it), list(map(squared, range(10))))
 
     def test_imap_unordered(self):
         it = self.pool.imap_unordered(sqr, range(1000))
-        self.assertEqual(sorted(it), map(sqr, range(1000)))
+        self.assertEqual(sorted(it), list(map(squared, range(1000))))
 
         it = self.pool.imap_unordered(sqr, range(1000))
-        self.assertEqual(sorted(it), map(sqr, range(1000)))
+        self.assertEqual(sorted(it), list(map(squared, range(1000))))
+
+    def test_imap_unordered_random(self):
+        it = self.pool.imap_unordered(sqr_random_sleep, range(10))
+        self.assertEqual(sorted(it), list(map(squared, range(10))))
+
+    def test_empty(self):
+        it = self.pool.imap_unordered(sqr, [])
+        self.assertEqual(list(it), [])
+
+        it = self.pool.imap(sqr, [])
+        self.assertEqual(list(it), [])
+
+        self.assertEqual(self.pool.map(sqr, []), [])
 
     def test_terminate(self):
-        result = self.pool.map_async(gevent.sleep, [0.1] * 1000)
+        result = self.pool.map_async(gevent.sleep, [0.1] * ((self.size or 10) * 2))
+        gevent.sleep(0.1)
         kill = TimingWrapper(self.pool.kill)
-        kill(block=True)
+        kill()
         assert kill.elapsed < 0.5, kill.elapsed
+        result.join()
+
+    def sleep(self, x):
+        gevent.sleep(float(x) / 10.)
+        return str(x)
+
+    def test_imap_unordered_sleep(self):
+        # testing that imap_unordered returns items in competion order
+        result = list(self.pool.imap_unordered(self.sleep, [10, 1, 2]))
+        if self.pool.size == 1:
+            expected = ['10', '1', '2']
+        else:
+            expected = ['1', '2', '10']
+        self.assertEqual(result, expected)
 
 
 class TestPool2(TestPool):
@@ -270,16 +352,22 @@ class TestJoinSleep(greentest.GenericWaitTestCase):
 
     def wait(self, timeout):
         p = pool.Pool()
-        p.spawn(gevent.sleep, 10)
-        p.join(timeout=timeout)
+        g = p.spawn(gevent.sleep, 10)
+        try:
+            p.join(timeout=timeout)
+        finally:
+            g.kill()
 
 
 class TestJoinSleep_raise_error(greentest.GenericWaitTestCase):
 
     def wait(self, timeout):
         p = pool.Pool()
-        p.spawn(gevent.sleep, 10)
-        p.join(timeout=timeout, raise_error=True)
+        g = p.spawn(gevent.sleep, 10)
+        try:
+            p.join(timeout=timeout, raise_error=True)
+        finally:
+            g.kill()
 
 
 class TestJoinEmpty(greentest.TestCase):
@@ -304,5 +392,29 @@ class TestSpawn(greentest.TestCase):
         self.assertEqual(len(p), 0)
 
 
+def error_iter():
+    yield 1
+    yield 2
+    raise ExpectedException
+
+
+class TestErrorInIterator(greentest.TestCase):
+    error_fatal = False
+
+    def test(self):
+        p = pool.Pool(3)
+        self.assertRaises(ExpectedException, p.map, lambda x: None, error_iter())
+        gevent.sleep(0.001)
+
+    def test_unordered(self):
+        p = pool.Pool(3)
+
+        def unordered():
+            return list(p.imap_unordered(lambda x: None, error_iter()))
+
+        self.assertRaises(ExpectedException, unordered)
+        gevent.sleep(0.001)
+
+
 if __name__ == '__main__':
     greentest.main()
