@@ -23,6 +23,7 @@ monkey.patch_all(thread=False)
 import cgi
 import os
 import sys
+import StringIO
 try:
     from wsgiref.validate import validator
 except ImportError:
@@ -33,6 +34,7 @@ except ImportError:
 import greentest
 import gevent
 from gevent import socket
+from gevent.pywsgi import Input
 
 
 CONTENT_LENGTH = 'Content-Length'
@@ -278,7 +280,8 @@ class CommonTests(TestCase):
                 fd.close()
             finally:
                 timeout.cancel()
-        except AssertionError, ex:
+        except AssertionError:
+            ex = sys.exc_info()[1]
             if ex is not exception:
                 raise
 
@@ -294,8 +297,9 @@ class CommonTests(TestCase):
         try:
             result = fd.readline()
             assert not result, 'The remote side is expected to close the connection, but it send %r' % (result, )
-        except socket.error, ex:
-            if ex[0] not in CONN_ABORTED_ERRORS:
+        except socket.error:
+            ex = sys.exc_info()[1]
+            if ex.args[0] not in CONN_ABORTED_ERRORS:
                 raise
 
     def SKIP_test_006_reject_long_urls(self):
@@ -626,22 +630,44 @@ class TestInputReadlines(TestInputReadline):
         return lines
 
 
-class TestError(TestCase):
+class TestInputN(TestCase):
+    # testing for this:
+    # File "/home/denis/work/gevent/gevent/pywsgi.py", line 70, in _do_read
+    #   if length and length > self.content_length - self.position:
+    # TypeError: unsupported operand type(s) for -: 'NoneType' and 'int'
 
-    @staticmethod
-    def application(env, start_response):
-        raise greentest.ExpectedException('TestError.application')
+    validator = None
+
+    def application(self, environ, start_response):
+        environ['wsgi.input'].read(5)
+        start_response('200 OK', [])
+        return []
 
     def test(self):
+        self.urlopen()
+
+
+class TestError(TestCase):
+
+    error = greentest.ExpectedException('TestError.application')
+    error_fatal = False
+
+    def application(self, env, start_response):
+        raise self.error
+
+    def test(self):
+        self.expect_one_error()
         self.urlopen(code=500)
+        self.assert_error(greentest.ExpectedException, self.error)
 
 
 class TestError_after_start_response(TestError):
 
-    @staticmethod
-    def application(env, start_response):
+    error = greentest.ExpectedException('TestError_after_start_response.application')
+
+    def application(self, env, start_response):
         start_response('200 OK', [('Content-Type', 'text/plain')])
-        raise greentest.ExpectedException('TestError_after_start_response.application')
+        raise self.error
 
 
 class TestEmptyYield(TestCase):
@@ -767,10 +793,7 @@ class ChunkedInputTests(TestCase):
         if dirt is None:
             dirt = self.dirt
 
-        b = ""
-        for c in chunks:
-            b += "%x%s\r\n%s\r\n" % (len(c), dirt, c)
-        return b
+        return chunk_encode(chunks, dirt=dirt)
 
     def body(self, dirt=None):
         return self.chunk_encode(["this", " is ", "chunked", "\nline", " 2", "\n", "line3", ""], dirt=dirt)
@@ -827,7 +850,8 @@ class ChunkedInputTests(TestCase):
         fd.write(req)
         try:
             read_http(fd, body="pong")
-        except AssertionError, ex:
+        except AssertionError:
+            ex = sys.exc_info()[1]
             if str(ex).startswith('Unexpected code: 400'):
                 if not server_implements_chunked:
                     print 'ChunkedNotImplementedWarning'
@@ -845,7 +869,8 @@ class ChunkedInputTests(TestCase):
         read_http(fd, body='this is chunked\nline 2\nline3')
 
     def test_close_before_finished(self):
-        self.hook_stderr()
+        if server_implements_chunked:
+            self.expect_one_error()
         body = '4\r\nthi'
         req = "POST /short-read HTTP/1.1\r\ntransfer-encoding: Chunked\r\n\r\n" + body
         fd = self.connect().makefile(bufsize=1)
@@ -853,18 +878,22 @@ class ChunkedInputTests(TestCase):
         fd.close()
         gevent.sleep(0.01)
         if server_implements_chunked:
-            self.assert_stderr_traceback(IOError, 'unexpected end of file while parsing chunked data')
+            self.assert_error(IOError, 'unexpected end of file while parsing chunked data')
 
 
 class Expect100ContinueTests(TestCase):
     validator = None
 
     def application(self, environ, start_response):
-        if int(environ['CONTENT_LENGTH']) > 1024:
+        content_length = int(environ['CONTENT_LENGTH'])
+        if content_length > 1024:
             start_response('417 Expectation Failed', [('Content-Length', '7'), ('Content-Type', 'text/plain')])
             return ['failure']
         else:
-            text = environ['wsgi.input'].read()
+            # pywsgi did sent a "100 continue" for each read
+            # see http://code.google.com/p/gevent/issues/detail?id=93
+            text = environ['wsgi.input'].read(1)
+            text += environ['wsgi.input'].read(content_length - 1)
             start_response('200 OK', [('Content-Length', str(len(text))), ('Content-Type', 'text/plain')])
             return [text]
 
@@ -874,7 +903,8 @@ class Expect100ContinueTests(TestCase):
         fd.write('PUT / HTTP/1.1\r\nHost: localhost\r\nContent-length: 1025\r\nExpect: 100-continue\r\n\r\n')
         try:
             read_http(fd, code=417, body="failure")
-        except AssertionError, ex:
+        except AssertionError:
+            ex = sys.exc_info()[1]
             if str(ex).startswith('Unexpected code: 400'):
                 if not server_implements_100continue:
                     print '100ContinueNotImplementedWarning'
@@ -1000,6 +1030,97 @@ class TestSubclass1(TestCase):
         self.assertEqual(fd.read(), '')
 
 
+class TestErrorAfterChunk(TestCase):
+    validator = None
+
+    @staticmethod
+    def application(env, start_response):
+        start_response('200 OK', [('Content-Type', 'text/plain')])
+        yield "hello"
+        raise greentest.ExpectedException('TestErrorAfterChunk')
+
+    def test(self):
+        fd = self.connect().makefile(bufsize=1)
+        self.expect_one_error()
+        fd.write('GET / HTTP/1.1\r\nHost: localhost\r\nConnection: keep-alive\r\n\r\n')
+        self.assertRaises(ValueError, read_http, fd)
+        self.assert_error(greentest.ExpectedException)
+
+
+def chunk_encode(chunks, dirt=None):
+    if dirt is None:
+        dirt = ""
+
+    b = ""
+    for c in chunks:
+        b += "%x%s\r\n%s\r\n" % (len(c), dirt, c)
+    return b
+
+
+class TestInputRaw(greentest.BaseTestCase):
+    def make_input(self, data, content_length=None, chunked_input=False):
+        if isinstance(data, list):
+            data = chunk_encode(data)
+            chunked_input = True
+
+        return Input(StringIO.StringIO(data), content_length=content_length, chunked_input=chunked_input)
+
+    def test_short_post(self):
+        i = self.make_input("1", content_length=2)
+        self.assertRaises(IOError, i.read)
+
+    def test_short_post_read_with_length(self):
+        i = self.make_input("1", content_length=2)
+        self.assertRaises(IOError, i.read, 2)
+
+    def test_short_post_readline(self):
+        i = self.make_input("1", content_length=2)
+        self.assertRaises(IOError, i.readline)
+
+    def test_post(self):
+        i = self.make_input("12", content_length=2)
+        data = i.read()
+        self.assertEqual(data, "12")
+
+    def test_post_read_with_length(self):
+        i = self.make_input("12", content_length=2)
+        data = i.read(10)
+        self.assertEqual(data, "12")
+
+    def test_chunked(self):
+        i = self.make_input(["1", "2", ""])
+        data = i.read()
+        self.assertEqual(data, "12")
+
+    def test_chunked_read_with_length(self):
+        i = self.make_input(["1", "2", ""])
+        data = i.read(10)
+        self.assertEqual(data, "12")
+
+    def test_chunked_missing_chunk(self):
+        i = self.make_input(["1", "2"])
+        self.assertRaises(IOError, i.read)
+
+    def test_chunked_missing_chunk_read_with_length(self):
+        i = self.make_input(["1", "2"])
+        self.assertRaises(IOError, i.read, 10)
+
+    def test_chunked_missing_chunk_readline(self):
+        i = self.make_input(["1", "2"])
+        self.assertRaises(IOError, i.readline)
+
+    def test_chunked_short_chunk(self):
+        i = self.make_input("2\r\n1", chunked_input=True)
+        self.assertRaises(IOError, i.read)
+
+    def test_chunked_short_chunk_read_with_length(self):
+        i = self.make_input("2\r\n1", chunked_input=True)
+        self.assertRaises(IOError, i.read, 2)
+
+    def test_chunked_short_chunk_readline(self):
+        i = self.make_input("2\r\n1", chunked_input=True)
+        self.assertRaises(IOError, i.readline)
+
 del CommonTests
 
 if __name__ == '__main__':
