@@ -1,4 +1,4 @@
-#!/usr/bin/python
+#!/usr/bin/env python
 """Unit test runner.
 
 This test runner runs each test module isolated in a subprocess, thus allowing them to
@@ -51,11 +51,13 @@ import platform
 
 try:
     import sqlite3
-except ImportError, ex:
+except ImportError:
+    ex = sys.exc_info()[1]
     sys.stderr.write('Failed to import sqlite3: %s\n' % ex)
     try:
         import pysqlite2.dbapi2 as sqlite3
-    except ImportError, ex:
+    except ImportError:
+        ex = sys.exc_info()[1]
         sys.stderr.write('Failed to import pysqlite2.dbapi2: %s\n' % ex)
         sqlite3 = None
 
@@ -75,7 +77,8 @@ def store_record(database_path, table, dictionary, _added_colums_per_db={}):
                 conn.execute(sql)
                 conn.commit()
                 _added_columns.add(key)
-            except sqlite3.OperationalError, ex:
+            except sqlite3.OperationalError:
+                ex = sys.exc_info()[1]
                 if 'duplicate column' not in str(ex).lower():
                     raise
     sql = 'insert or replace into %s (%s) values (%s)' % (table, ', '.join(keys), ', '.join(':%s' % key for key in keys))
@@ -83,7 +86,24 @@ def store_record(database_path, table, dictionary, _added_colums_per_db={}):
     try:
         cursor.execute(sql, dictionary)
     except sqlite3.Error:
-        print 'sql=%r\ndictionary=%r' % (sql, dictionary)
+        print ('sql=%r\ndictionary=%r' % (sql, dictionary))
+        raise
+    conn.commit()
+    return cursor.lastrowid
+
+
+def delete_record(database_path, table, dictionary, _added_colums_per_db={}):
+    if sqlite3 is None:
+        return
+    keys = dictionary.keys()
+    conn = sqlite3.connect(database_path)
+    print ('deleting %s from database' % (dictionary, ))
+    sql = 'delete from %s where %s' % (table, ' AND '.join('%s=:%s' % (key, key) for key in keys))
+    cursor = conn.cursor()
+    try:
+        cursor.execute(sql, dictionary)
+    except sqlite3.Error:
+        print ('sql=%r\ndictionary=%r' % (sql, dictionary))
         raise
     conn.commit()
     return cursor.lastrowid
@@ -108,6 +128,10 @@ class DatabaseTestResult(_TextTestResult):
         from time import time
         self.time = time()
 
+    def addSkip(self, test, reason):
+        delete_record(self.database_path, 'testcase', self.params)
+        return super(DatabaseTestResult, self).addSkip(test, reason)
+
     def _store_result(self, test, result):
         self.params['result'] = result
         from time import time
@@ -162,12 +186,9 @@ def get_changeset():
     return changeset
 
 
-def get_libevent_version():
+def get_core_version():
     from gevent import core
-    libevent_version = core.get_version()
-    if core.get_header_version() != core.get_version() and core.get_header_version() is not None:
-        libevent_version += '/headers=%s' % core.get_header_version()
-    return libevent_version
+    return core.get_version()
 
 
 def get_tempnam():
@@ -180,6 +201,20 @@ def get_tempnam():
     return os.path.join(os.path.dirname(tempnam), 'testresults.sqlite3')
 
 
+def execfile_as_main(path):
+    import __builtin__
+    oldmain = sys.modules["__main__"]
+    main = sys.__class__("__main__")
+    main.__file__ = path
+    main.__builtins__ = __builtin__
+    main.__package__ = None
+    try:
+        sys.modules["__main__"] = main
+        return execfile(path, main.__dict__)
+    finally:
+        sys.modules["__main__"] = oldmain
+
+
 def run_tests(options, args):
     arg = args[0]
     module_name = arg
@@ -192,16 +227,21 @@ def run_tests(options, args):
             return DatabaseTestRunner(database_path=options.db, runid=options.runid, module_name=module_name, verbosity=options.verbosity)
 
     if options.db:
+        try:
+            from unittest import runner
+        except ImportError:
+            pass
+        else:
+            runner.TextTestRunner = _runner
         import unittest
         unittest.TextTestRunner = _runner
         import test_support
         test_support.BasicTestRunner = _runner
 
     sys.argv = args
-    globals()['__file__'] = arg
 
     if os.path.exists(arg):
-        execfile(arg, globals())
+        execfile_as_main(arg)
     else:
         test = defaultTestLoader.loadTestsFromName(arg)
         result = _runner().run(test)
@@ -228,7 +268,7 @@ def run_subprocess(args, options):
 
     def killer():
         retcode.append('TIMEOUT')
-        print >> sys.stderr, 'Killing %s (%s) because of timeout' % (popen.pid, args)
+        sys.stderr.write('Killing %s (%s) because of timeout\n' % (popen.pid, args))
         popen.kill()
 
     timeout = Timer(options.timeout, killer)
@@ -274,22 +314,24 @@ def spawn_subprocess(args, options, base_params):
         params['id'] = row_id
     retcode, output, output_printed = run_subprocess(args, options)
     if len(output) > OUTPUT_LIMIT:
-        output = output[:OUTPUT_LIMIT] + '<AbridgedOutputWarning>'
+        warn = '<AbridgedOutputWarning>'
+        output = output[:OUTPUT_LIMIT - len(warn)] + warn
     if retcode:
         if retcode == 1 and 'test_support.TestSkipped' in output:
             pass
         else:
             if not output_printed and options.verbosity >= -1:
                 sys.stdout.write(output)
-            print '%s failed with code %s' % (' '.join(args), retcode)
+            print ('%s failed with code %s' % (' '.join(args), retcode))
     elif retcode == 0:
         if not output_printed and options.verbosity >= 1:
             sys.stdout.write(output)
         if options.verbosity >= 0:
-            print '%s passed' % ' '.join(args)
+            print ('%s passed' % ' '.join(args))
         success = True
     else:
-        print '%s timed out' % ' '.join(args)
+        print ('%s timed out' % ' '.join(args))
+    sys.stdout.flush()
     if options.db:
         params['output'] = output
         params['retcode'] = retcode
@@ -301,7 +343,7 @@ def spawn_subprocesses(options, args):
     params = {'runid': options.runid,
               'python': '%s.%s.%s' % sys.version_info[:3],
               'changeset': get_changeset(),
-              'libevent_version': get_libevent_version(),
+              'core_version': get_core_version(),
               'uname': platform.uname()[0],
               'retcode': 'TIMEOUT'}
     success = True
@@ -321,12 +363,12 @@ def spawn_subprocesses(options, args):
             traceback.print_exc()
     if options.db:
         try:
-            print '-' * 80
+            print ('-' * 80)
             if print_stats(options):
                 success = False
         except sqlite3.OperationalError:
             traceback.print_exc()
-        print 'To view stats again for this run, use %s --stats --runid %s --db %s' % (sys.argv[0], options.runid, options.db)
+        print ('To view stats again for this run, use %s --stats --runid %s --db %s' % (sys.argv[0], options.runid, options.db))
     if not success:
         sys.exit(1)
 
@@ -354,8 +396,6 @@ def get_failed_testcases(cursor, runid):
 
 
 _warning_re = re.compile('\w*warning', re.I)
-_error_re = re.compile(r'(?P<prefix>\s*)Traceback \(most recent call last\):' +
-                       r'(\n(?P=prefix)[ \t]+[^\n]*)+\n(?P=prefix)(?P<error>[\w\.]+)')
 
 
 def get_warnings(output):
@@ -377,7 +417,19 @@ def get_exceptions(output):
     ... ZeroDivisionError: integer division or modulo by zero''')
     ['ZeroDivisionError']
     """
-    return [x.group('error') for x in _error_re.finditer(output)]
+    errors = []
+    readtb = False
+    for line in output.split('\n'):
+        if 'Traceback (most recent call last):' in line:
+            readtb = True
+        else:
+            if readtb:
+                if line[:1] == ' ':
+                    pass
+                else:
+                    errors.append(line.split(':')[0])
+                    readtb = False
+    return errors
 
 
 def get_warning_stats(output):
@@ -386,7 +438,9 @@ def get_warning_stats(output):
         counter.setdefault(warning, 0)
         counter[warning] += 1
     items = counter.items()
-    items.sort(key=lambda (a, b): -b)
+    def sortkey(x):
+        return -x[1]
+    items.sort(key=sortkey)
     result = []
     for name, count in items:
         if count == 1:
@@ -418,7 +472,9 @@ def get_traceback_stats(output, test):
             counter[error] += 1
         traceback_count -= 1
     items = counter.items()
-    items.sort(key=lambda (a, b): -b)
+    def sortkey(x):
+        return -x[1]
+    items.sort(key=sortkey)
     if traceback_count > 0:
         items.append(('other traceback', traceback_count))
     result = []
@@ -431,7 +487,7 @@ def get_traceback_stats(output, test):
 
 
 def get_info(output, test):
-    output = output[:OUTPUT_LIMIT * 2]
+    output = output[:OUTPUT_LIMIT]
     traceback_stats, ignored_list = get_traceback_stats(output, test)
     warning_stats = get_warning_stats(output)
     result = traceback_stats + warning_stats
@@ -444,14 +500,14 @@ def print_stats(options):
     cursor = db.cursor()
     if options.runid is None:
         options.runid = cursor.execute('select runid from test order by started_at desc limit 1').fetchall()[0][0]
-        print 'Using the latest runid: %s' % options.runid
+        print ('Using the latest runid: %s' % options.runid)
     total = len(get_testcases(cursor, options.runid))
     failed, errors = get_failed_testcases(cursor, options.runid)
     timedout = get_testcases(cursor, options.runid, 'TIMEOUT')
     for test, output, retcode in cursor.execute('select test, output, retcode from test where runid=?', (options.runid, )):
         info, skipped = get_info(output or '', test)
         if info:
-            print '%s: %s' % (test, info)
+            print ('%s: %s' % (test, info))
         if retcode == 'TIMEOUT':
             for testcase in timedout:
                 if testcase.startswith(test + '.'):
@@ -469,18 +525,18 @@ def print_stats(options):
                     total += 1
     if failed:
         failed.sort()
-        print 'FAILURES: '
+        print ('FAILURES: ')
         for testcase in failed:
             error = errors.get(testcase)
             if error:
                 error = repr(error)[1:-1][:100]
-                print ' - %s: %s' % (testcase, error)
+                print (' - %s: %s' % (testcase, error))
             else:
-                print ' - %s' % (testcase, )
+                print (' - %s' % (testcase, ))
     if timedout:
-        print 'TIMEOUTS: '
-        print ' - ' + '\n - '.join(timedout)
-    print '%s testcases passed; %s failed; %s timed out' % (total, len(failed), len(timedout))
+        print ('TIMEOUTS: ')
+        print (' - ' + '\n - '.join(timedout))
+    print ('%s testcases passed; %s failed; %s timed out' % (total, len(failed), len(timedout)))
     if failed or timedout:
         return True
     return False
@@ -505,10 +561,11 @@ def main():
 
     if options.db:
         if sqlite3:
-            print 'Using the database: %s' % options.db
+            options.db = os.path.abspath(options.db)
+            print ('Using the database: %s' % options.db)
         else:
-            sys.exit('Cannot access the database %r: no sqlite3 module found.' % (options.db, ))
-        options.db = os.path.abspath(options.db)
+            sys.stderr.write('Cannot access the database %r: no sqlite3 module found.\n' % (options.db, ))
+            options.db = False
 
     if options.db:
         db = sqlite3.connect(options.db)
@@ -526,7 +583,7 @@ def main():
             except ImportError:
                 import random
                 options.runid = str(random.random())[2:]
-            print 'Generated runid: %s' % (options.runid, )
+            print ('Generated runid: %s' % (options.runid, ))
         if options.record:
             run_tests(options, args)
         else:
